import base64
import binascii
import codecs
import contextlib
import hashlib
import os
import random

import wandb
from wandb import env
from wandb import util
from wandb.data_types import WBValue

if wandb.TYPE_CHECKING:  # type: ignore

    from typing import (
        List,
        Optional,
        Union,
        Dict,
        Callable,
        TYPE_CHECKING,
        Sequence,
        Tuple,
    )

    if TYPE_CHECKING:
        import wandb.filesync.step_prepare.StepPrepare as StepPrepare  # type: ignore


def md5_string(string):
    hash_md5 = hashlib.md5()
    hash_md5.update(string.encode())
    return base64.b64encode(hash_md5.digest()).decode("ascii")


def b64_string_to_hex(string):
    return binascii.hexlify(base64.standard_b64decode(string)).decode("ascii")


def md5_hash_file(path):
    hash_md5 = hashlib.md5()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(64 * 1024), b""):
            hash_md5.update(chunk)
    return hash_md5


def md5_file_b64(path):
    return base64.b64encode(md5_hash_file(path).digest()).decode("ascii")


def md5_file_hex(path):
    return md5_hash_file(path).hexdigest()


def bytes_to_hex(bytestr):
    # Works in python2 / python3
    return codecs.getencoder("hex")(bytestr)[0]


class ArtifactManifest(object):
    # entries: Dict[str, "ArtifactEntry"]

    @classmethod
    # TODO: we don't need artifact here.
    def from_manifest_json(cls, artifact, manifest_json):
        if "version" not in manifest_json:
            raise ValueError("Invalid manifest format. Must contain version field.")
        version = manifest_json["version"]
        for sub in cls.__subclasses__():
            if sub.version() == version:
                return sub.from_manifest_json(artifact, manifest_json)

    @classmethod
    def version(cls):
        pass

    def __init__(self, artifact, storage_policy, entries=None):
        self.artifact = artifact
        self.storage_policy = storage_policy
        self.entries = entries or {}

    def to_manifest_json(self):
        raise NotImplementedError()

    def digest(self):
        raise NotImplementedError()

    def add_entry(self, entry):
        if (
            entry.path in self.entries
            and entry.digest != self.entries[entry.path].digest
        ):
            raise ValueError("Cannot add the same path twice: %s" % entry.path)
        self.entries[entry.path] = entry

    def get_entry_by_path(self, path):
        return self.entries.get(path)

    def get_entries_in_directory(self, directory):
        return [
            self.entries[entry_key]
            for entry_key in self.entries
            if entry_key.startswith(
                directory + "/"
            )  # entries use forward slash even for windows
        ]


class ArtifactEntry(object):
    # path: str
    # ref: Optional[str]
    # digest: str
    # birth_artifact_id: Optional[str]
    # size: Optional[int]
    # extra: Dict
    # local_path: Optional[str]

    def parent_artifact(self):
        """
        Get the artifact to which this artifact entry belongs.

        Returns:
            (Artifact): The parent artifact
        """
        raise NotImplementedError

    def download(self, root = None):
        """
        Downloads this artifact entry to the specified root path.

        Arguments:
            root: (str, optional) The root path in which to download this
                artifact entry. Defaults to the artifact's root.

        Returns:
            (str): The path of the downloaded artifact entry.

        """
        raise NotImplementedError

    def ref_target(self):
        """
        Gets the reference URL that this artifact entry targets.
        Returns:
            (str): The reference URL of this artifact entry.
        Raises:
            ValueError: If this artifact entry was not a reference.
        """
        raise NotImplementedError

    def ref_url(self):
        """
        Gets a URL to this artifact entry such that it can be referenced
        by another artifact.

        Returns:
            (str): A URL representing this artifact entry.

        Examples:
            Basic usage
            ```
            ref_url = source_artifact.get_path('file.txt').ref_url()
            derived_artifact.add_reference(ref_url)
            ```
        """
        raise NotImplementedError


class Artifact(object):
    @property
    def id(self):
        """
        Returns:
            (str): The artifact's ID
        """
        raise NotImplementedError

    @property
    def version(self):
        """
        Returns:
            (int): The version of this artifact. For example, if this
                is the first version of an artifact, its `version` will
                be 'v0'.
        """
        raise NotImplementedError

    @property
    def name(self):
        """
        Returns:
            (str): The artifact's name
        """
        raise NotImplementedError

    @property
    def type(self):
        """
        Returns:
            (str): The artifact's type
        """
        raise NotImplementedError

    @property
    def entity(self):
        """
        Returns:
            (str): The name of the entity this artifact belongs to.
        """
        raise NotImplementedError

    @property
    def project(self):
        """
        Returns:
            (str): The name of the project this artifact belongs to.
        """
        raise NotImplementedError

    @property
    def manifest(self):
        """
        Returns:
            (ArtifactManifest): The artifact's manifest, listing all of its contents.
                You cannot add more files to an artifact once you've retrieved its
                manifest.
        """
        raise NotImplementedError

    @property
    def digest(self):
        """
        Returns:
            (str): The artifact's logical digest, a checksum of its contents. If
                an artifact has the same digest as the current `latest` version,
                then `log_artifact` is a no-op.
        """
        raise NotImplementedError

    @property
    def state(self):
        """
        Returns:
            (str): The state of the artifact, which can be one of "PENDING",
                "COMMITTED", or "DELETED".
        """
        raise NotImplementedError

    @property
    def size(self):
        """
        Returns:
            (int): The size in bytes of the artifact. Includes any references
                tracked by this artifact.
        """
        raise NotImplementedError

    @property
    def commit_hash(self):
        """
        Returns:
            (str): The artifact's commit hash which is used in http URLs
        """
        raise NotImplementedError

    @property
    def description(self):
        """
        Returns:
            (str): Free text that offers a description of the artifact. The
                description is markdown rendered in the UI, so this is a good place
                to put links, etc.
        """
        raise NotImplementedError

    @description.setter
    def description(self, desc):
        """
        Arguments:
            desc: Free text that offers a description of the artifact. The
                description is markdown rendered in the UI, so this is a good place
                to put links, etc.
        """
        raise NotImplementedError

    @property
    def metadata(self):
        """
        Returns:
            (dict): Structured data associated with the artifact,
                for example class distribution of a dataset. This will eventually be queryable
                and plottable in the UI. There is a hard limit of 100 total keys.
        """
        raise NotImplementedError

    @metadata.setter
    def metadata(self, metadata):
        """
        Arguments:
            metadata: (dict) Structured data associated with the artifact,
                for example class distribution of a dataset. This will eventually be queryable
                and plottable in the UI. There is a hard limit of 100 total keys.
        """
        raise NotImplementedError

    @property
    def aliases(self):
        """
        Returns:
            (list): A list of the aliases associated with this artifact. The list is
                mutable and calling `save()` will persist all alias changes.
        """
        raise NotImplementedError

    @aliases.setter
    def aliases(self, aliases):
        """
        Arguments:
            aliases: (list) The list of aliases associated with this artifact.
        """
        raise NotImplementedError

    def used_by(self):
        """
        Returns:
            (list): A list of the runs that have used this artifact.
        """
        raise NotImplementedError

    def logged_by(self):
        """
        Returns:
            (Run): The run that first logged this artifact.
        """
        raise NotImplementedError

    def new_file(self, name, mode = "w"):
        """
        Open a new temporary file that will be automatically added to the artifact.

        Arguments:
            name: (str) The name of the new file being added to the artifact.
            mode: (str, optional) The mode in which to open the new file.

        Examples:
            ```
            artifact = wandb.Artifact('my_data', type='dataset')
            with artifact.new_file('hello.txt') as f:
                f.write('hello!')
            wandb.log_artifact(artifact)
            ```

        Returns:
            (file): A new file object that can be written to. Upon closing,
                the file will be automatically added to the artifact.
        """
        raise NotImplementedError

    def add_file(
        self,
        local_path,
        name = None,
        is_tmp = False,
    ):
        """
        Adds a local file to the artifact.

        Arguments:
            local_path: (str) The path to the file being added.
            name: (str, optional) The path within the artifact to use for the file being added. Defaults
                to the basename of the file.
            is_tmp: (bool, optional) If true, then the file is renamed deterministically to avoid collisions.
                (default: False)

        Examples:
            Adding a file without an explicit name:
            ```
            artifact.add_file('path/to/file.txt') # Added as `file.txt'
            ```

            Adding a file with an explicit name:
            ```
            artifact.add_file('path/to/file.txt', name='new/path/file.txt') # Added as 'new/path/file.txt'
            ```

        Raises:
            Exception: if problem

        Returns:
            ArtifactManifestEntry: the added manifest entry

        """
        raise NotImplementedError

    def add_dir(self, local_path, name = None):
        """
        Adds a local directory to the artifact.

        Arguments:
            local_path: (str) The path to the directory being added.
            name: (str, optional) The path within the artifact to use for the directory being added. Defaults
                to files being added under the root of the artifact.

        Examples:
            Adding a directory without an explicit name:
            ```
            artifact.add_dir('my_dir/') # All files in `my_dir/` are added at the root of the artifact.
            ```

            Adding a directory without an explicit name:
            ```
            artifact.add_dir('my_dir/', path='destination') # All files in `my_dir/` are added under `destination/`.
            ```

        Raises:
            Exception: if problem.

        Returns:
            None
        """
        raise NotImplementedError

    def add_reference(
        self,
        uri,
        name = None,
        checksum = True,
        max_objects = None,
    ):
        """
        Adds a reference denoted by a URI to the artifact. Unlike adding files or directories,
        references are NOT uploaded to W&B. However, artifact methods such as `download()` can
        be used regardless of whether the artifact contains references or uploaded files.

        By default, W&B offers special
        handling for the following schemes:

        - http(s): The size and digest of the file will be inferred by the `Content-Length` and
            the `ETag` response headers returned by the server.
        - s3: The checksum and size will be pulled from the object metadata. If bucket versioning
            is enabled, then the version ID is also tracked.
        - gs: The checksum and size will be pulled from the object metadata. If bucket versioning
            is enabled, then the version ID is also tracked.
        - file: The checksum and size will be pulled from the file system. This scheme is useful if
            you have an NFS share or other externally mounted volume containing files you wish to track
            but not necessarily upload.

        For any other scheme, the digest is just a hash of the URI and the size is left blank.

        Arguments:
            uri: (str) The URI path of the reference to add. Can be an object returned from
                Artifact.get_path to store a reference to another artifact's entry.
            name: (str) The path within the artifact to place the contents of this reference
            checksum: (bool, optional) Whether or not to checksum the resource(s) located at the
                reference URI. Checksumming is strongly recommended as it enables automatic integrity
                validation, however it can be disabled to speed up artifact creation. (default: True)
            max_objects: (int, optional) The maximum number of objects to consider when adding a
                reference that points to directory or bucket store prefix. For S3 and GCS, this limit
                is 10,000 by default but is uncapped for other URI schemes. (default: None)

        Raises:
            Exception: If problem.

        Returns:
            List[ArtifactManifestEntry]: The added manifest entries.

        Examples:
            Adding an HTTP link:
            ```
            # Adds `file.txt` to the root of the artifact as a reference
            artifact.add_reference('http://myserver.com/file.txt')
            ```

            Adding an S3 prefix without an explicit name:
            ```
            # All objects under `prefix/` will be added at the root of the artifact.
            artifact.add_reference('s3://mybucket/prefix')
            ```

            Adding a GCS prefix with an explicit name:
            ```
            # All objects under `prefix/` will be added under `path/` at the top of the artifact.
            artifact.add_reference('gs://mybucket/prefix', name='path')
            ```
        """
        raise NotImplementedError

    def add(self, obj, name):
        """Adds wandb.WBValue `obj` to the artifact.

        ```
        obj = artifact.get(name)
        ```

        Arguments:
            obj: (wandb.WBValue) The object to add. Currently support one of
                Bokeh, JoinedTable, PartitionedTable, Table, Classes, ImageMask,
                BoundingBoxes2D, Audio, Image, Video, Html, Object3D
            name: (str) The path within the artifact to add the object.

        Returns:
            ArtifactManifestEntry: the added manifest entry

        Examples:
            Basic usage
            ```
            artifact = wandb.Artifact('my_table', 'dataset')
            table = wandb.Table(columns=["a", "b", "c"], data=[[i, i*2, 2**i]])
            artifact.add(table, "my_table")

            wandb.log_artifact(artifact)
            ```

            Retrieving an object:
            ```
            artifact = wandb.use_artifact('my_table:latest')
            table = artifact.get("my_table")
            ```
        """
        raise NotImplementedError

    def get_path(self, name):
        """
        Gets the path to the file located at the artifact relative `name`.

        NOTE: This will raise an error unless the artifact has been fetched using
        `use_artifact`, fetched using the API, or `wait()` has been called.

        Arguments:
            name: (str) The artifact relative name to get

        Raises:
            Exception: if problem

        Examples:
            Basic usage
            ```
            # Run logging the artifact
            with wandb.init() as r:
                artifact = wandb.Artifact('my_dataset', type='dataset')
                artifact.add_file('path/to/file.txt')
                wandb.log_artifact(artifact)

            # Run using the artifact
            with wandb.init() as r:
                artifact = r.use_artifact('my_dataset:latest')
                path = artifact.get_path('file.txt')

                # Can now download 'file.txt' directly:
                path.download()
            ```
        """
        raise NotImplementedError

    def get(self, name):
        """
        Gets the WBValue object located at the artifact relative `name`.

        NOTE: This will raise an error unless the artifact has been fetched using
        `use_artifact`, fetched using the API, or `wait()` has been called.

        Arguments:
            name: (str) The artifact relative name to get

        Raises:
            Exception: if problem

        Examples:
            Basic usage
            ```
            # Run logging the artifact
            with wandb.init() as r:
                artifact = wandb.Artifact('my_dataset', type='dataset')
                table = wandb.Table(columns=["a", "b", "c"], data=[[i, i*2, 2**i]])
                artifact.add(table, "my_table")
                wandb.log_artifact(artifact)

            # Run using the artifact
            with wandb.init() as r:
                artifact = r.use_artifact('my_dataset:latest')
                table = r.get('my_table')
            ```
        """
        raise NotImplementedError

    def download(self, root = None, recursive = False):
        """
        Downloads the contents of the artifact to the specified root directory.

        NOTE: Any existing files at `root` are left untouched. Explicitly delete
        root before calling `download` if you want the contents of `root` to exactly
        match the artifact.

        Arguments:
            root: (str, optional) The directory in which to download this artifact's files.
            recursive: (bool, optional) If true, then all dependent artifacts are eagerly
                downloaded. Otherwise, the dependent artifacts are downloaded as needed.

        Returns:
            (str): The path to the downloaded contents.
        """
        raise NotImplementedError

    def checkout(self, root = None):
        """
        Replaces the specified root directory with the contents of the artifact.

        WARNING: This will DELETE all files in `root` that are not included in the
        artifact.

        Arguments:
            root: (str, optional) The directory to replace with this artifact's files.

        Returns:
           (str): The path to the checked out contents.
        """
        raise NotImplementedError

    def verify(self, root = None):
        """
        Verify that the actual contents of an artifact at a specified directory
        `root` match the expected contents of the artifact according to its
        manifest.

        All files in the directory are checksummed and the checksums are then
        cross-referenced against the artifact's manifest.

        NOTE: References are not verified.

        Arguments:
            root: (str, optional) The directory to verify. If None
                artifact will be downloaded to './artifacts/<self.name>/'

        Raises:
            (ValueError): If the verification fails.
        """
        raise NotImplementedError

    def save(self):
        """
        Persists any changes made to the artifact.

        Returns:
            None
        """
        raise NotImplementedError

    def delete(self):
        """
        Deletes this artifact, cleaning up all files associated with it.

        NOTE: Deletion is permanent and CANNOT be undone.

        Returns:
            None
        """
        raise NotImplementedError

    def wait(self):
        """
        Waits for this artifact to finish logging, if needed.

        Returns:
            Artifact
        """
        raise NotImplementedError

    def __getitem__(self, name):
        """
        Gets the WBValue object located at the artifact relative `name`.

        NOTE: This will raise an error unless the artifact has been fetched using
        `use_artifact`, fetched using the API, or `wait()` has been called.

        Arguments:
            name: (str) The artifact relative name to get

        Raises:
            Exception: if problem

        Examples:
            Basic usage
            ```
            artifact = wandb.Artifact('my_table', 'dataset')
            table = wandb.Table(columns=["a", "b", "c"], data=[[i, i*2, 2**i]])
            artifact["my_table"] = table

            wandb.log_artifact(artifact)
            ```

            Retrieving an object:
            ```
            artifact = wandb.use_artifact('my_table:latest')
            table = artifact["my_table"]
            ```
        """
        raise NotImplementedError

    def __setitem__(self, name, item):
        """
        Adds `item` to the artifact at path `name`

        Arguments:
            name: (str) The path within the artifact to add the object.
            item: (wandb.WBValue) The object to add.

        Returns:
            ArtifactManifestEntry: the added manifest entry

        Examples:
            Basic usage
            ```
            artifact = wandb.Artifact('my_table', 'dataset')
            table = wandb.Table(columns=["a", "b", "c"], data=[[i, i*2, 2**i]])
            artifact["my_table"] = table

            wandb.log_artifact(artifact)
            ```

            Retrieving an object:
            ```
            artifact = wandb.use_artifact('my_table:latest')
            table = artifact["my_table"]
            ```
        """
        raise NotImplementedError


class StorageLayout(object):
    V1 = "V1"
    V2 = "V2"


class StoragePolicy(object):
    @classmethod
    def lookup_by_name(cls, name):
        for sub in cls.__subclasses__():
            if sub.name() == name:
                return sub
        return None

    @classmethod
    def name(cls):
        pass

    @classmethod
    def from_config(cls, config):
        pass

    def config(self):
        pass

    def load_file(
        self, artifact, name, manifest_entry
    ):
        raise NotImplementedError

    def store_file(
        self,
        artifact_id,
        artifact_manifest_id,
        entry,
        preparer,
        progress_callback = None,
    ):
        raise NotImplementedError

    def store_reference(
        self, artifact, path, name=None, checksum=True, max_objects=None
    ):
        raise NotImplementedError

    def load_reference(
        self,
        artifact,
        name,
        manifest_entry,
        local = False,
    ):
        raise NotImplementedError


class StorageHandler(object):
    @property
    def scheme(self):
        """
        :return: The scheme to which this handler applies.
        :rtype: str
        """
        pass

    def load_path(
        self, artifact, manifest_entry, local = False,
    ):
        """
        Loads the file or directory within the specified artifact given its
        corresponding index entry.

        :param manifest_entry: The index entry to load
        :type manifest_entry: ArtifactManifestEntry
        :return: A path to the file represented by `index_entry`
        :rtype: str
        """
        pass

    def store_path(
        self, artifact, path, name=None, checksum=True, max_objects=None
    ):
        """
        Stores the file or directory at the given path within the specified artifact.

        :param path: The path to store
        :type path: str
        :param name: If specified, the logical name that should map to `path`
        :type name: str
        :return: A list of manifest entries to store within the artifact
        :rtype: list(ArtifactManifestEntry)
        """
        pass


class ArtifactsCache(object):

    _TMP_PREFIX = "tmp"

    def __init__(self, cache_dir):
        self._cache_dir = cache_dir
        util.mkdir_exists_ok(self._cache_dir)
        self._md5_obj_dir = os.path.join(self._cache_dir, "obj", "md5")
        self._etag_obj_dir = os.path.join(self._cache_dir, "obj", "etag")
        self._artifacts_by_id = {}
        self._random = random.Random()
        self._random.seed()

    def check_md5_obj_path(self, b64_md5, size):
        hex_md5 = util.bytes_to_hex(base64.b64decode(b64_md5))
        path = os.path.join(self._cache_dir, "obj", "md5", hex_md5[:2], hex_md5[2:])
        opener = self._cache_opener(path)
        if os.path.isfile(path) and os.path.getsize(path) == size:
            return path, True, opener
        util.mkdir_exists_ok(os.path.dirname(path))
        return path, False, opener

    def check_etag_obj_path(self, etag, size):
        path = os.path.join(self._cache_dir, "obj", "etag", etag[:2], etag[2:])
        opener = self._cache_opener(path)
        if os.path.isfile(path) and os.path.getsize(path) == size:
            return path, True, opener
        util.mkdir_exists_ok(os.path.dirname(path))
        return path, False, opener

    def get_artifact(self, artifact_id):
        return self._artifacts_by_id.get(artifact_id)

    def store_artifact(self, artifact):
        self._artifacts_by_id[artifact.id] = artifact

    def cleanup(self, target_size):
        bytes_reclaimed = 0
        paths = {}
        total_size = 0
        for root, _, files in os.walk(self._cache_dir):
            for file in files:
                path = os.path.join(root, file)
                stat = os.stat(path)

                if file.startswith(ArtifactsCache._TMP_PREFIX):
                    try:
                        os.remove(path)
                        bytes_reclaimed += stat.st_size
                    except OSError:
                        pass
                    continue

                paths[path] = stat
                total_size += stat.st_size

        sorted_paths = sorted(paths.items(), key=lambda x: x[1].st_atime)
        for path, stat in sorted_paths:
            if total_size < target_size:
                return bytes_reclaimed

            try:
                os.remove(path)
            except OSError:
                pass

            total_size -= stat.st_size
            bytes_reclaimed += stat.st_size
        return bytes_reclaimed

    def _cache_opener(self, path):
        @contextlib.contextmanager
        def helper(mode="w"):
            dirname = os.path.dirname(path)
            tmp_file = os.path.join(
                dirname,
                "%s_%s"
                % (
                    ArtifactsCache._TMP_PREFIX,
                    util.rand_alphanumeric(length=8, rand=self._random),
                ),
            )
            with util.fsync_open(tmp_file, mode=mode) as f:
                yield f

            try:
                # Use replace where we can, as it implements an atomic
                # move on most platforms. If it doesn't exist, we have
                # to use rename which isn't atomic in all cases but there
                # isn't a better option.
                #
                # The atomic replace is important in the event multiple processes
                # attempt to write to / read from the cache at the same time. Each
                # writer firsts stages its writes to a temporary file in the cache.
                # Once it is finished, we issue an atomic replace operation to update
                # the cache. Although this can result in redundant downloads, this
                # guarantees that readers can NEVER read incomplete files from the
                # cache.
                #
                # IMPORTANT: Replace is NOT atomic across different filesystems. This why
                # it is critical that the temporary files sit directly in the cache --
                # they need to be on the same filesystem!
                os.replace(tmp_file, path)
            except AttributeError:
                os.rename(tmp_file, path)

        return helper


_artifacts_cache = None


def get_artifacts_cache():
    global _artifacts_cache
    if _artifacts_cache is None:
        cache_dir = os.path.join(env.get_cache_dir(), "artifacts")
        _artifacts_cache = ArtifactsCache(cache_dir)
    return _artifacts_cache



#

import wandb
from wandb.proto.wandb_telemetry_pb2 import Imports as TelemetryImports
from wandb.proto.wandb_telemetry_pb2 import TelemetryRecord

if wandb.TYPE_CHECKING:  # type: ignore
    from typing import ContextManager, Type, Optional
    from types import TracebackType

    # avoid cycle, use string type reference
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from .. import wandb_run


class _TelemetryObject(object):
    _run: Optional["wandb_run.Run"]

    def __init__(self, run: "wandb_run.Run" = None) -> None:
        self._run = run or wandb.run
        self._obj = TelemetryRecord()

    def __enter__(self) -> TelemetryRecord:
        return self._obj

    def __exit__(
        self,
        exctype: Optional[Type[BaseException]],
        excinst: Optional[BaseException],
        exctb: Optional[TracebackType],
    ) -> None:
        if not self._run:
            return
        self._run._telemetry_callback(self._obj)


def context(run: "wandb_run.Run" = None) -> ContextManager[TelemetryRecord]:
    return _TelemetryObject(run=run)


__all__ = [
    "TelemetryImports",
    "TelemetryRecord",
    "context",
]



from copy import copy

import six

from . import ast
from .visitor_meta import QUERY_DOCUMENT_KEYS, VisitorMeta


class Falsey(object):

    def __nonzero__(self):
        return False

    def __bool__(self):
        return False


BREAK = object()
REMOVE = Falsey()


class Stack(object):
    __slots__ = 'in_array', 'index', 'keys', 'edits', 'prev'

    def __init__(self, in_array, index, keys, edits, prev):
        self.in_array = in_array
        self.index = index
        self.keys = keys
        self.edits = edits
        self.prev = prev


def visit(root, visitor, key_map=None):
    visitor_keys = key_map or QUERY_DOCUMENT_KEYS

    stack = None
    in_array = isinstance(root, list)
    keys = [root]
    index = -1
    edits = []
    parent = None
    path = []
    ancestors = []
    new_root = root
    leave = visitor.leave
    enter = visitor.enter
    path_pop = path.pop
    ancestors_pop = ancestors.pop
    path_append = path.append
    ancestors_append = ancestors.append

    while True:
        index += 1
        is_leaving = index == len(keys)
        is_edited = is_leaving and edits
        if is_leaving:
            key = path_pop() if ancestors else None
            node = parent
            parent = ancestors_pop() if ancestors else None

            if is_edited:
                if in_array:
                    node = list(node)

                else:
                    node = copy(node)
                edit_offset = 0
                for edit_key, edit_value in edits:
                    if in_array:
                        edit_key -= edit_offset

                    if in_array and edit_value is REMOVE:
                        node.pop(edit_key)
                        edit_offset += 1

                    else:
                        if isinstance(node, list):
                            node[edit_key] = edit_value

                        else:
                            setattr(node, edit_key, edit_value)

            index = stack.index
            keys = stack.keys
            edits = stack.edits
            in_array = stack.in_array
            stack = stack.prev

        else:
            if parent:
                key = index if in_array else keys[index]
                if isinstance(parent, list):
                    node = parent[key]

                else:
                    node = getattr(parent, key, None)

            else:
                key = None
                node = new_root

            if node is REMOVE or node is None:
                continue

            if parent:
                path_append(key)

        result = None

        if not isinstance(node, list):
            assert isinstance(node, ast.Node), 'Invalid AST Node: ' + repr(node)

            if is_leaving:
                result = leave(node, key, parent, path, ancestors)

            else:
                result = enter(node, key, parent, path, ancestors)

            if result is BREAK:
                break

            if result is False:
                if not is_leaving:
                    path_pop()
                    continue

            elif result is not None:
                edits.append((key, result))
                if not is_leaving:
                    if isinstance(result, ast.Node):
                        node = result

                    else:
                        path_pop()
                        continue

        if result is None and is_edited:
            edits.append((key, node))

        if not is_leaving:
            stack = Stack(in_array, index, keys, edits, stack)
            in_array = isinstance(node, list)
            keys = node if in_array else visitor_keys.get(type(node), None) or []
            index = -1
            edits = []

            if parent:
                ancestors_append(parent)

            parent = node

        if not stack:
            break

    if edits:
        new_root = edits[-1][1]

    return new_root


@six.add_metaclass(VisitorMeta)
class Visitor(object):
    __slots__ = ()

    def enter(self, node, key, parent, path, ancestors):
        method = self._get_enter_handler(type(node))
        if method:
            return method(self, node, key, parent, path, ancestors)

    def leave(self, node, key, parent, path, ancestors):
        method = self._get_leave_handler(type(node))
        if method:
            return method(self, node, key, parent, path, ancestors)


class ParallelVisitor(Visitor):
    __slots__ = 'skipping', 'visitors'

    def __init__(self, visitors):
        self.visitors = visitors
        self.skipping = [None] * len(visitors)

    def enter(self, node, key, parent, path, ancestors):
        for i, visitor in enumerate(self.visitors):
            if not self.skipping[i]:
                result = visitor.enter(node, key, parent, path, ancestors)
                if result is False:
                    self.skipping[i] = node
                elif result is BREAK:
                    self.skipping[i] = BREAK
                elif result is not None:
                    return result

    def leave(self, node, key, parent, path, ancestors):
        for i, visitor in enumerate(self.visitors):
            if not self.skipping[i]:
                result = visitor.leave(node, key, parent, path, ancestors)
                if result is BREAK:
                    self.skipping[i] = BREAK
                elif result is not None and result is not False:
                    return result
            elif self.skipping[i] == node:
                self.skipping[i] = REMOVE


class TypeInfoVisitor(Visitor):
    __slots__ = 'visitor', 'type_info'

    def __init__(self, type_info, visitor):
        self.type_info = type_info
        self.visitor = visitor

    def enter(self, node, key, parent, path, ancestors):
        self.type_info.enter(node)
        result = self.visitor.enter(node, key, parent, path, ancestors)
        if result is not None:
            self.type_info.leave(node)
            if isinstance(result, ast.Node):
                self.type_info.enter(result)
        return result

    def leave(self, node, key, parent, path, ancestors):
        result = self.visitor.leave(node, key, parent, path, ancestors)
        self.type_info.leave(node)
        return result



#
# -*- coding: utf-8 -*-
"""InternalRun - Internal-only run object.

Semi-stubbed run for internal process use.

"""
from wandb._globals import _datatypes_set_callback

from .. import wandb_run


class InternalRun(wandb_run.Run):
    def __init__(self, run_obj, settings, datatypes_cb):
        super(InternalRun, self).__init__(settings=settings)
        self._run_obj = run_obj

        # TODO: This overwrites what's done in the constructor of wandb_run.Run.
        # We really want a common interface for wandb_run.Run and InternalRun.
        _datatypes_set_callback(datatypes_cb)

    def _set_backend(self, backend):
        # This type of run object can't have a backend
        # or do any writes.
        pass



#
import json

import wandb


if wandb.TYPE_CHECKING:
    from typing import Any, Dict, Union
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:  # pragma: no cover
        from wandb.proto import wandb_internal_pb2 as pb
        from wandb.proto import wandb_telemetry_pb2 as tpb


def dict_from_proto_list(obj_list):
    d = dict()
    for item in obj_list:
        d[item.key] = json.loads(item.value_json)
    return d


def proto_encode_to_dict(
    pb_obj: Union["tpb.TelemetryRecord", "pb.MetricRecord"]
) -> Dict[int, Any]:
    data: Dict[int, Any] = dict()
    fields = pb_obj.ListFields()
    for desc, value in fields:
        if desc.name.startswith("_"):
            continue
        if desc.type == desc.TYPE_STRING:
            data[desc.number] = value
        elif desc.type == desc.TYPE_INT32:
            data[desc.number] = value
        elif desc.type == desc.TYPE_ENUM:
            data[desc.number] = value
        elif desc.type == desc.TYPE_MESSAGE:
            nested = value.ListFields()
            bool_msg = all(d.type == d.TYPE_BOOL for d, _ in nested)
            if bool_msg:
                items = [d.number for d, v in nested if v]
                if items:
                    data[desc.number] = items
    return data



import json
import os
import socket

from . import files as sm_files


def parse_sm_secrets():
    """We read our api_key from secrets.env in SageMaker"""
    env_dict = dict()
    # Set secret variables
    if os.path.exists(sm_files.SM_SECRETS):
        for line in open(sm_files.SM_SECRETS, "r"):
            key, val = line.strip().split("=", 1)
            env_dict[key] = val
    return env_dict


def parse_sm_resources():
    run_dict = dict()
    env_dict = dict()
    run_id = os.getenv("TRAINING_JOB_NAME")
    if run_id:
        run_dict["run_id"] = "-".join(
            [run_id, os.getenv("CURRENT_HOST", socket.gethostname())]
        )
    conf = json.load(open(sm_files.SM_RESOURCE_CONFIG))
    if len(conf["hosts"]) > 1:
        run_dict["group"] = os.getenv("TRAINING_JOB_NAME")
    env_dict = parse_sm_secrets()
    return run_dict, env_dict



import numpy as np
from wandb.sweeps import random_search
import random
import pytest


sweep_config_2params = {'parameters': {
    'v1': {'min': 3, 'max': 5},
    'v2': {'min': 5, 'max': 6}}}


sweep_config_bad = {'parameters': {
    'v1': {'max': 3, 'min': 5},
    'v2': {'min': 5, 'max': 6}}}


def test_rand_single():
    random.seed(73)
    np.random.seed(73)
    gs = random_search.RandomSearch()
    runs = []
    sweep = {'config': sweep_config_2params, 'runs': runs}
    params, info = gs.next_run(sweep)
    assert info == None
    assert params['v1']['value'] == 3 and params['v2']['value'] == 6


def test_rand_bad():
    random.seed(73)
    np.random.seed(73)
    gs = random_search.RandomSearch()
    runs = []
    sweep = {'config': sweep_config_bad, 'runs': runs}
    with pytest.raises(ValueError) as excinfo:
        params, info = gs.next_run(sweep)



#
import json
import logging
import os

import six
from wandb.errors import Error
from wandb.util import load_yaml
import yaml

from . import filesystem


logger = logging.getLogger("wandb")


class ConfigError(Error):  # type: ignore
    pass


def dict_from_proto_list(obj_list):
    d = dict()
    for item in obj_list:
        d[item.key] = dict(desc=None, value=json.loads(item.value_json))
    return d


def update_from_proto(config_dict, config_proto):
    for item in config_proto.update:
        key_list = item.nested_key or (item.key,)
        assert key_list, "key or nested key must be set"
        target = config_dict
        # recurse down the dictionary structure:
        for prop in key_list[:-1]:
            if not target.get(prop):
                target[prop] = {}
            target = target[prop]
        # use the last element of the key to write the leaf:
        target[key_list[-1]] = json.loads(item.value_json)
    for item in config_proto.remove:
        key_list = item.nested_key or (item.key,)
        assert key_list, "key or nested key must be set"
        target = config_dict
        # recurse down the dictionary structure:
        for prop in key_list[:-1]:
            target = target[prop]
        # use the last element of the key to write the leaf:
        del target[key_list[-1]]
        # TODO(jhr): should we delete empty parents?


def dict_add_value_dict(config_dict):
    d = dict()
    for k, v in six.iteritems(config_dict):
        d[k] = dict(desc=None, value=v)
    return d


def dict_strip_value_dict(config_dict):
    d = dict()
    for k, v in six.iteritems(config_dict):
        d[k] = v["value"]
    return d


def dict_no_value_from_proto_list(obj_list):
    d = dict()
    for item in obj_list:
        possible_dict = json.loads(item.value_json)
        if not isinstance(possible_dict, dict) or "value" not in possible_dict:
            # (tss) TODO: This is protecting against legacy 'wandb_version' field.
            # Should investigate why the config payload even has 'wandb_version'.
            logger.warning("key '{}' has no 'value' attribute".format(item.key))
            continue
        d[item.key] = possible_dict["value"]

    return d


# TODO(jhr): these functions should go away once we merge jobspec PR
def save_config_file_from_dict(config_filename, config_dict):
    s = b"wandb_version: 1"
    if config_dict:  # adding an empty dictionary here causes a parse error
        s += b"\n\n" + yaml.dump(
            config_dict,
            Dumper=yaml.SafeDumper,
            default_flow_style=False,
            allow_unicode=True,
            encoding="utf-8",
        )
    data = s.decode("utf-8")
    filesystem._safe_makedirs(os.path.dirname(config_filename))
    with open(config_filename, "w") as conf_file:
        conf_file.write(data)


def dict_from_config_file(filename, must_exist=False):
    if not os.path.exists(filename):
        if must_exist:
            raise ConfigError("config file %s doesn't exist" % filename)
        logger.debug("no default config file found in %s" % filename)
        return None
    try:
        conf_file = open(filename)
    except OSError:
        raise ConfigError("Couldn't read config file: %s" % filename)
    try:
        loaded = load_yaml(conf_file)
    except yaml.parser.ParserError:
        raise ConfigError("Invalid YAML in config yaml")
    config_version = loaded.pop("wandb_version", None)
    if config_version is not None and config_version != 1:
        raise ConfigError("Unknown config version")
    data = dict()
    for k, v in six.iteritems(loaded):
        data[k] = v["value"]
    return data



"""
Base classes to be inherited from for Search and EarlyTerminate algorithms
"""

import math

from wandb.sweeps.util import is_nan_or_nan_string


class Search():
    def _metric_from_run(self, sweep_config, run, default=None):
        metric = None
        metric_name = sweep_config['metric']['name']

        maximize = False
        if 'goal' in sweep_config['metric']:
            if sweep_config['metric']['goal'] == 'maximize':
                maximize = True

        # Use summary to find metric
        if metric_name in run.summaryMetrics:
            metric = run.summaryMetrics[metric_name]
            # Exclude None or NaN summary metrics.
            if metric is None or is_nan_or_nan_string(metric):
                metric = None
        if maximize and metric is not None:
            metric = -metric

        # Use history to find metric (if available)
        metric_history = []
        run_history = getattr(run, 'history', [])
        for line in run_history:
            m = line.get(metric_name)
            if m is None:
                continue
            if is_nan_or_nan_string(m):
                continue
            metric_history.append(m)
        if maximize:
            metric_history = [-m for m in metric_history]

        # find minimum from summary or history
        if metric_history:
            if metric:
                metric_history.append(metric)
            metric = min(metric_history)

        # use default if specified
        if metric is None:
            if default is None:
                raise ValueError(
                    "Couldn't find summary metric {}".format(metric_name))
            metric = default
        return metric

    def next_run(self, sweep):
        """Called each time an agent requests new work.
        Arguments:
            sweep: <defined above>
        Returns:
            None if all work complete for this sweep. A dictionary of configuration
            parameters for the next run.
        """
        raise NotImplementedError


class EarlyTerminate():
    def _load_metric_name_and_goal(self, sweep_config):
        if not 'metric' in sweep_config:
            raise ValueError("Key 'metric' required for early termination")

        self.metric_name = sweep_config['metric']['name']

        self.maximize = False
        if 'goal' in sweep_config['metric']:
            if sweep_config['metric']['goal'] == 'maximize':
                self.maximize = True

    def _load_run_metric_history(self, run):
        metric_history = []
        for line in run.history:
            if self.metric_name in line:
                m = line[self.metric_name]
                metric_history.append(m)

        # Filter out bad values
        metric_history = [
            x for x in metric_history
            if x is not None and not is_nan_or_nan_string(x)
        ]
        if self.maximize:
            metric_history = [-m for m in metric_history]

        return metric_history

    def stop_runs(self, sweep_config, runs):
        return [], {}



#!/usr/bin/env python
import time
import wandb
import random
import argparse

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument("--project", type=str, default=None, help="project")
parser.add_argument("--epochs", type=int, default=27, help="epochs")
parser.add_argument("--variance", type=float, default=1, help="param1")
parser.add_argument("--base", type=float, default=5, help="param2")
parser.add_argument("--increment", type=float, default=5, help="param2")
parser.add_argument("--direction", type=float, default=5, help="param2")
parser.add_argument("--sleep", type=float, default=5, help="param2")
args = parser.parse_args()

run = wandb.init(project=args.project)
# Force configuration update (should not be needed in the future)
wandb.config.update({'dummy': 1})

time.sleep(1)

metric = args.base + random.random()*args.variance
for e, x in enumerate(range(args.epochs)):
    metric += args.increment * args.direction
    if metric < 0:
        metric = 0
    wandb.log({"loss_metric": metric})
    print("INFO:  %3d = %f" % (e, metric))
    time.sleep(args.sleep)



import wandb


def histogram(table, value, title=None):
    """
    Construct a histogram plot.

    Arguments:
        table (wandb.Table): Table of data.
        label (string): Name of column to use as data for bucketing.
        title (string): Plot title.

    Returns:
        A plot object, to be passed to wandb.log()

    Example:
        ```
        data = [[i, random.random() + math.sin(i / 10)] for i in range(100)]
        table = wandb.Table(data=data, columns=["step", "height"])
        wandb.log({'histogram-plot1': wandb.plot.histogram(table, "height")})
        ```
    """
    return wandb.plot_table(
        "wandb/histogram/v0", table, {"value": value}, {"title": title}
    )



"""
Grid Search
"""

import itertools
import random
from wandb.sweeps.params import HyperParameter, HyperParameterSet
from wandb.sweeps.base import Search


class GridSearch(Search):
    def __init__(self, randomize_order=False):
        self.randomize_order = randomize_order

    def next_run(self, sweep):
        if 'parameters' not in sweep['config']:
            raise ValueError('Grid search requires "parameters" section')
        config = sweep['config']['parameters']
        params = HyperParameterSet.from_config(config)

        # Check that all parameters are categorical or constant
        for p in params:
            if p.type != HyperParameter.CATEGORICAL and p.type != HyperParameter.CONSTANT:
                raise ValueError(
                    'Parameter %s is a disallowed type with grid search. Grid search requires all parameters to be categorical or constant' % p.name)

        # we can only deal with discrete params in a grid search
        discrete_params = [p for p in params if p.type ==
                           HyperParameter.CATEGORICAL]

        # build an iterator over all combinations of param values
        param_names = [p.name for p in discrete_params]
        param_values = [p.values for p in discrete_params]
        param_value_set = list(itertools.product(*param_values))

        if self.randomize_order:
            random.shuffle(param_value_set)

        new_value_set = next(
            (value_set for value_set in param_value_set
             # check if parameter set is contained in some run
                if not self._runs_contains_param_values(sweep['runs'], dict(zip(param_names, value_set))
                                                        )
             ), None)

        # handle the case where we couldn't find a unique parameter set
        if new_value_set == None:
            return None

        # set next_run_params based on our new set of params
        for param, value in zip(discrete_params, new_value_set):
            param.value = value

        return (params.to_config(), None)

    def _run_contains_param_values(self, run, params):
        for key, value in params.items():
            if not key in run.config:
                return False
            if not run.config[key]['value'] == value:
                #print("not same {} {}".format(run.config[key], value))
                return False
        return True

    def _runs_contains_param_values(self, runs, params):
        ret_val = any(self._run_contains_param_values(run, params)
                      for run in runs)
        return any(self._run_contains_param_values(run, params) for run in runs)



# File is generated by: tox -e codemod
# -*- coding: utf-8 -*-
"""Handle Manager."""

from __future__ import print_function

import json
import logging
import math
import numbers
import os

import six
import wandb
from wandb.proto import wandb_internal_pb2

from . import meta, sample, stats
from . import tb_watcher
from ..lib import handler_util, proto_util


if wandb.TYPE_CHECKING:
    from typing import (
        TYPE_CHECKING,
        Any,
        Callable,
        Dict,
        List,
        Tuple,
        Sequence,
        Iterable,
        Optional,
        cast,
    )
    from .settings_static import SettingsStatic
    from six.moves.queue import Queue
    from threading import Event
    from ..interface.interface import BackendSender
    from wandb.proto.wandb_internal_pb2 import Record, Result

    SummaryDict = Dict[str, Any]


logger = logging.getLogger(__name__)


def _dict_nested_set(target, key_list, v):
    # recurse down the dictionary structure:
    for k in key_list[:-1]:
        target.setdefault(k, {})
        new_target = target.get(k)
        if wandb.TYPE_CHECKING and TYPE_CHECKING:
            new_target = cast(Dict[str, Any], new_target)
        target = new_target
    # use the last element of the key to write the leaf:
    target[key_list[-1]] = v


class HandleManager(object):

    # _consolidated_summary: SummaryDict
    # _sampled_history: Dict[str, sample.UniformSampleAccumulator]
    # _settings: SettingsStatic
    # _record_q: "Queue[Record]"
    # _result_q: "Queue[Result]"
    # _stopped: Event
    # _sender_q: "Queue[Record]"
    # _writer_q: "Queue[Record]"
    # _interface: BackendSender
    # _system_stats: Optional[stats.SystemStats]
    # _tb_watcher: Optional[tb_watcher.TBWatcher]
    # _metric_defines: Dict[str, wandb_internal_pb2.MetricRecord]
    # _metric_globs: Dict[str, wandb_internal_pb2.MetricRecord]
    # _metric_track: Dict[Tuple[str, ...], float]
    # _metric_copy: Dict[Tuple[str, ...], Any]

    def __init__(
        self,
        settings,
        record_q,
        result_q,
        stopped,
        sender_q,
        writer_q,
        interface,
    ):
        self._settings = settings
        self._record_q = record_q
        self._result_q = result_q
        self._stopped = stopped
        self._sender_q = sender_q
        self._writer_q = writer_q
        self._interface = interface

        self._tb_watcher = None
        self._system_stats = None
        self._step = 0

        # keep track of summary from key/val updates
        self._consolidated_summary = dict()
        self._sampled_history = dict()
        self._metric_defines = dict()
        self._metric_globs = dict()
        self._metric_track = dict()
        self._metric_copy = dict()

    def handle(self, record):
        record_type = record.WhichOneof("record_type")
        assert record_type
        handler_str = "handle_" + record_type
        handler = getattr(self, handler_str, None)
        assert handler, "unknown handle: {}".format(handler_str)
        handler(record)

    def handle_request(self, record):
        request_type = record.request.WhichOneof("request_type")
        assert request_type
        handler_str = "handle_request_" + request_type
        handler = getattr(self, handler_str, None)
        if request_type != "network_status":
            logger.debug("handle_request: {}".format(request_type))
        assert handler, "unknown handle: {}".format(handler_str)
        handler(record)

    def _dispatch_record(self, record, always_send = False):
        if not self._settings._offline or always_send:
            self._sender_q.put(record)
        if not record.control.local:
            self._writer_q.put(record)

    def handle_request_defer(self, record):
        defer = record.request.defer
        state = defer.state

        logger.info("handle defer: {}".format(state))
        # only handle flush tb (sender handles the rest)
        if state == defer.FLUSH_STATS:
            if self._system_stats:
                # TODO(jhr): this could block so we dont really want to call shutdown
                # from handler thread
                self._system_stats.shutdown()
        elif state == defer.FLUSH_TB:
            if self._tb_watcher:
                # shutdown tensorboard workers so we get all metrics flushed
                self._tb_watcher.finish()
                self._tb_watcher = None
        elif state == defer.FLUSH_SUM:
            self._save_summary(self._consolidated_summary, flush=True)

        # defer is used to drive the sender finish state machine
        self._dispatch_record(record, always_send=True)

    def handle_request_login(self, record):
        self._dispatch_record(record)

    def handle_run(self, record):
        self._dispatch_record(record)

    def handle_stats(self, record):
        self._dispatch_record(record)

    def handle_config(self, record):
        self._dispatch_record(record)

    def handle_output(self, record):
        self._dispatch_record(record)

    def handle_files(self, record):
        self._dispatch_record(record)

    def handle_artifact(self, record):
        self._dispatch_record(record)

    def handle_alert(self, record):
        self._dispatch_record(record)

    def _save_summary(self, summary_dict, flush = False):
        summary = wandb_internal_pb2.SummaryRecord()
        for k, v in six.iteritems(summary_dict):
            update = summary.update.add()
            update.key = k
            update.value_json = json.dumps(v)
        record = wandb_internal_pb2.Record(summary=summary)
        if flush:
            self._dispatch_record(record)
        elif not self._settings._offline:
            self._sender_q.put(record)

    def _save_history(self, record):
        for item in record.history.item:
            # TODO(jhr) save nested keys?
            k = item.key
            v = json.loads(item.value_json)
            if isinstance(v, numbers.Real):
                self._sampled_history.setdefault(k, sample.UniformSampleAccumulator())
                self._sampled_history[k].add(v)

    def _update_summary_metrics(
        self,
        s,
        kl,
        v,
        float_v,
        goal_max,
    ):
        updated = False
        best_key = None
        if s.none:
            return False
        if s.copy:
            # non key list copy already done in _update_summary
            if len(kl) > 1:
                _dict_nested_set(self._consolidated_summary, kl, v)
                return True
        if s.last:
            last_key = tuple(kl + ["last"])
            old_last = self._metric_track.get(last_key)
            if old_last is None or float_v != old_last:
                self._metric_track[last_key] = float_v
                _dict_nested_set(self._consolidated_summary, last_key, v)
                updated = True
        if s.best:
            best_key = tuple(kl + ["best"])
        if s.max or best_key and goal_max:
            max_key = tuple(kl + ["max"])
            old_max = self._metric_track.get(max_key)
            if old_max is None or float_v > old_max:
                self._metric_track[max_key] = float_v
                if s.max:
                    _dict_nested_set(self._consolidated_summary, max_key, v)
                    updated = True
                if best_key:
                    _dict_nested_set(self._consolidated_summary, best_key, v)
                    updated = True
        # defaulting to minimize if goal is not supecified
        if s.min or best_key and not goal_max:
            min_key = tuple(kl + ["min"])
            old_min = self._metric_track.get(min_key)
            if old_min is None or float_v < old_min:
                self._metric_track[min_key] = float_v
                if s.min:
                    _dict_nested_set(self._consolidated_summary, min_key, v)
                    updated = True
                if best_key:
                    _dict_nested_set(self._consolidated_summary, best_key, v)
                    updated = True
        if s.mean:
            tot_key = tuple(kl + ["tot"])
            num_key = tuple(kl + ["num"])
            avg_key = tuple(kl + ["mean"])
            tot = self._metric_track.get(tot_key, 0.0)
            num = self._metric_track.get(num_key, 0)
            tot += float_v
            num += 1
            self._metric_track[tot_key] = tot
            self._metric_track[num_key] = num
            _dict_nested_set(self._consolidated_summary, avg_key, tot / num)
            updated = True
        return updated

    def _update_summary_leaf(
        self,
        kl,
        v,
        d = None,
    ):
        has_summary = d and d.HasField("summary")
        if len(kl) == 1:
            copy_key = tuple(kl)
            old_copy = self._metric_copy.get(copy_key)
            if old_copy is None or v != old_copy:
                self._metric_copy[copy_key] = v
                # Store copy metric if not specified, or copy behavior
                if not has_summary or (d and d.summary.copy):
                    self._consolidated_summary[kl[0]] = v
                    return True
        if not d:
            return False
        if not has_summary:
            return False
        if not isinstance(v, numbers.Real):
            return False
        if math.isnan(v):
            return False
        float_v = float(v)
        goal_max = None
        if d.goal:
            goal_max = d.goal == d.GOAL_MAXIMIZE
        if self._update_summary_metrics(
            d.summary, kl=kl, v=v, float_v=float_v, goal_max=goal_max
        ):
            return True
        return False

    def _update_summary_list(
        self,
        kl,
        v,
        d = None,
    ):
        metric_key = ".".join([k.replace(".", "\\.") for k in kl])
        d = self._metric_defines.get(metric_key, d)
        # if the dict has _type key, its a wandb table object
        if isinstance(v, dict) and not handler_util.metric_is_wandb_dict(v):
            updated = False
            for nk, nv in six.iteritems(v):
                if self._update_summary_list(kl=kl[:] + [nk], v=nv, d=d):
                    updated = True
            return updated
        updated = self._update_summary_leaf(kl=kl, v=v, d=d)
        return updated

    def _update_summary(self, history_dict):
        # keep old behavior fast path if no define metrics have been used
        if not self._metric_defines:
            self._consolidated_summary.update(history_dict)
            return True
        updated = False
        for k, v in six.iteritems(history_dict):
            if self._update_summary_list(kl=[k], v=v):
                updated = True
        return updated

    def _history_assign_step(self, record, history_dict):
        has_step = record.history.HasField("step")
        item = record.history.item.add()
        item.key = "_step"
        if has_step:
            step = record.history.step.num
            history_dict["_step"] = step
            item.value_json = json.dumps(step)
            self._step = step + 1
        else:
            history_dict["_step"] = self._step
            item.value_json = json.dumps(self._step)
            self._step += 1

    def _history_define_metric(
        self, hkey
    ):
        """check for hkey match in glob metrics, return defined metric."""
        # Dont define metric for internal metrics
        if hkey.startswith("_"):
            return None
        for k, mglob in six.iteritems(self._metric_globs):
            if k.endswith("*"):
                if hkey.startswith(k[:-1]):
                    m = wandb_internal_pb2.MetricRecord()
                    m.CopyFrom(mglob)
                    m.ClearField("glob_name")
                    m.options.defined = False
                    m.name = hkey
                    return m
        return None

    def _history_update_leaf(
        self, kl, v, history_dict, update_history
    ):
        hkey = ".".join([k.replace(".", "\\.") for k in kl])
        m = self._metric_defines.get(hkey)
        if not m:
            m = self._history_define_metric(hkey)
            if not m:
                return
            mr = wandb_internal_pb2.Record()
            mr.metric.CopyFrom(m)
            mr.control.local = True  # Dont store this, just send it
            self._handle_defined_metric(mr)

        if m.options.step_sync and m.step_metric:
            if m.step_metric not in history_dict:
                copy_key = tuple([m.step_metric])
                step = self._metric_copy.get(copy_key)
                if step is not None:
                    update_history[m.step_metric] = step

    def _history_update_list(
        self, kl, v, history_dict, update_history
    ):
        if isinstance(v, dict):
            for nk, nv in six.iteritems(v):
                self._history_update_list(
                    kl=kl[:] + [nk],
                    v=nv,
                    history_dict=history_dict,
                    update_history=update_history,
                )
            return
        self._history_update_leaf(
            kl=kl, v=v, history_dict=history_dict, update_history=update_history
        )

    def _history_update(self, record, history_dict):
        # if syncing an old run, we can skip this logic
        if history_dict.get("_step") is None:
            self._history_assign_step(record, history_dict)

        update_history = {}
        # Look for metric matches
        if self._metric_defines or self._metric_globs:
            for hkey, hval in six.iteritems(history_dict):
                self._history_update_list([hkey], hval, history_dict, update_history)

        if update_history:
            history_dict.update(update_history)
            for k, v in six.iteritems(update_history):
                item = record.history.item.add()
                item.key = k
                item.value_json = json.dumps(v)

    def handle_history(self, record):
        history_dict = proto_util.dict_from_proto_list(record.history.item)
        self._history_update(record, history_dict)
        self._dispatch_record(record)
        self._save_history(record)

        updated = self._update_summary(history_dict)
        if updated:
            self._save_summary(self._consolidated_summary)

    def handle_summary(self, record):
        summary = record.summary

        for item in summary.update:
            if len(item.nested_key) > 0:
                # we use either key or nested_key -- not both
                assert item.key == ""
                key = tuple(item.nested_key)
            else:
                # no counter-assertion here, because technically
                # summary[""] is valid
                key = (item.key,)

            target = self._consolidated_summary

            # recurse down the dictionary structure:
            for prop in key[:-1]:
                target = target[prop]

            # use the last element of the key to write the leaf:
            target[key[-1]] = json.loads(item.value_json)

        for item in summary.remove:
            if len(item.nested_key) > 0:
                # we use either key or nested_key -- not both
                assert item.key == ""
                key = tuple(item.nested_key)
            else:
                # no counter-assertion here, because technically
                # summary[""] is valid
                key = (item.key,)

            target = self._consolidated_summary

            # recurse down the dictionary structure:
            for prop in key[:-1]:
                target = target[prop]

            # use the last element of the key to erase the leaf:
            del target[key[-1]]

        self._save_summary(self._consolidated_summary)

    def handle_exit(self, record):
        self._dispatch_record(record, always_send=True)

    def handle_final(self, record):
        self._dispatch_record(record, always_send=True)

    def handle_header(self, record):
        self._dispatch_record(record)

    def handle_footer(self, record):
        self._dispatch_record(record)

    def handle_request_check_version(self, record):
        self._dispatch_record(record)

    def handle_request_log_artifact(self, record):
        self._dispatch_record(record)

    def handle_telemetry(self, record):
        self._dispatch_record(record)

    def handle_request_run_start(self, record):
        run_start = record.request.run_start
        assert run_start
        assert run_start.run

        if not self._settings._disable_stats:
            pid = os.getpid()
            self._system_stats = stats.SystemStats(pid=pid, interface=self._interface)
            self._system_stats.start()

        if not self._settings._disable_meta:
            run_meta = meta.Meta(settings=self._settings, interface=self._interface)
            run_meta.probe()
            run_meta.write()

        self._tb_watcher = tb_watcher.TBWatcher(
            self._settings, interface=self._interface, run_proto=run_start.run
        )

        if run_start.run.resumed:
            self._step = run_start.run.starting_step
        result = wandb_internal_pb2.Result(uuid=record.uuid)
        self._result_q.put(result)

    def handle_request_resume(self, record):
        if self._system_stats is not None:
            logger.info("starting system metrics thread")
            self._system_stats.start()

    def handle_request_pause(self, record):
        if self._system_stats is not None:
            logger.info("stopping system metrics thread")
            self._system_stats.shutdown()

    def handle_request_poll_exit(self, record):
        self._dispatch_record(record, always_send=True)

    def handle_request_stop_status(self, record):
        self._dispatch_record(record)

    def handle_request_network_status(self, record):
        self._dispatch_record(record)

    def handle_request_get_summary(self, record):
        result = wandb_internal_pb2.Result(uuid=record.uuid)
        for key, value in six.iteritems(self._consolidated_summary):
            item = wandb_internal_pb2.SummaryItem()
            item.key = key
            item.value_json = json.dumps(value)
            result.response.get_summary_response.item.append(item)
        self._result_q.put(result)

    def handle_tbrecord(self, record):
        logger.info("handling tbrecord: %s", record)
        if self._tb_watcher:
            tbrecord = record.tbrecord
            self._tb_watcher.add(tbrecord.log_dir, tbrecord.save, tbrecord.root_dir)
        self._dispatch_record(record)

    def _handle_defined_metric(self, record):
        metric = record.metric
        if metric._control.overwrite:
            self._metric_defines.setdefault(
                metric.name, wandb_internal_pb2.MetricRecord()
            ).CopyFrom(metric)
        else:
            self._metric_defines.setdefault(
                metric.name, wandb_internal_pb2.MetricRecord()
            ).MergeFrom(metric)

        # before dispatching, make sure step_metric is defined, if not define it and
        # dispatch it locally first
        metric = self._metric_defines[metric.name]
        if metric.step_metric and metric.step_metric not in self._metric_defines:
            m = wandb_internal_pb2.MetricRecord(name=metric.step_metric)
            self._metric_defines[metric.step_metric] = m
            mr = wandb_internal_pb2.Record()
            mr.metric.CopyFrom(m)
            mr.control.local = True  # Dont store this, just send it
            self._dispatch_record(mr)

        self._dispatch_record(record)

    def _handle_glob_metric(self, record):
        metric = record.metric
        if metric._control.overwrite:
            self._metric_globs.setdefault(
                metric.glob_name, wandb_internal_pb2.MetricRecord()
            ).CopyFrom(metric)
        else:
            self._metric_globs.setdefault(
                metric.glob_name, wandb_internal_pb2.MetricRecord()
            ).MergeFrom(metric)
        self._dispatch_record(record)

    def handle_metric(self, record):
        """Handle MetricRecord.

        Walkthrough of the life of a MetricRecord:

        Metric defined:
        - run.define_metric() parses arguments create wandb_metric.Metric
        - build MetricRecord publish to interface
        - handler (this function) keeps list of metrics published:
          - self._metric_defines: Fully defined metrics
          - self._metric_globs: metrics that have a wildcard
        - dispatch writer and sender thread
          - writer: records are saved to persistent store
          - sender: fully defined metrics get mapped into metadata for UI

        History logged:
        - handle_history
        - check if metric matches _metric_defines
        - if not, check if metric matches _metric_globs
        - if _metric globs match, generate defined metric and call _handle_metric

        Args:
            record (Record): Metric record to process
        """
        if record.metric.name:
            self._handle_defined_metric(record)
        elif record.metric.glob_name:
            self._handle_glob_metric(record)

    def handle_request_sampled_history(self, record):
        result = wandb_internal_pb2.Result(uuid=record.uuid)
        for key, sampled in six.iteritems(self._sampled_history):
            item = wandb_internal_pb2.SampledHistoryItem()
            item.key = key
            values = sampled.get()
            if all(isinstance(i, numbers.Integral) for i in values):
                item.values_int.extend(values)
            elif all(isinstance(i, numbers.Real) for i in values):
                item.values_float.extend(values)
            result.response.sampled_history_response.item.append(item)
        self._result_q.put(result)

    def handle_request_shutdown(self, record):
        # TODO(jhr): should we drain things and stop new requests from coming in?
        result = wandb_internal_pb2.Result(uuid=record.uuid)
        self._result_q.put(result)
        self._stopped.set()

    def finish(self):
        logger.info("shutting down handler")
        if self._tb_watcher:
            self._tb_watcher.finish()



"""
This module hooks fast.ai Learners to Weights & Biases through a callback.
Requested logged data can be configured through the callback constructor.

Examples:
    WandbCallback can be used when initializing the Learner::

    ```
        from wandb.fastai import WandbCallback
        [...]
        learn = Learner(data, ..., callback_fns=WandbCallback)
        learn.fit(epochs)
    ```

    Custom parameters can be given using functools.partial::

    ```
        from wandb.fastai import WandbCallback
        from functools import partial
        [...]
        learn = Learner(data, ..., callback_fns=partial(WandbCallback, ...))
        learn.fit(epochs)
    ```

    Finally, it is possible to use WandbCallback only when starting
    training. In this case it must be instantiated::

    ```
        learn.fit(..., callbacks=WandbCallback(learn))
    ```

    or, with custom parameters::

    ```
        learn.fit(..., callbacks=WandbCallback(learn, ...))
    ```
"""
import wandb
import fastai
from fastai.callbacks import TrackerCallback
from pathlib import Path
import random

try:
    import matplotlib

    if wandb.wandb_lib.ipython._get_python_type() != "jupyter":
        matplotlib.use("Agg")  # non-interactive backend (avoid tkinter issues)
    import matplotlib.pyplot as plt
except:
    print("Warning: matplotlib required if logging sample image predictions")


class WandbCallback(TrackerCallback):
    """
    Automatically saves model topology, losses & metrics.
    Optionally logs weights, gradients, sample predictions and best trained model.

    Arguments:
        learn (fastai.basic_train.Learner): the fast.ai learner to hook.
        log (str): "gradients", "parameters", "all", or None. Losses & metrics are always logged.
        save_model (bool): save model at the end of each epoch. It will also load best model at the end of training.
        monitor (str): metric to monitor for saving best model. None uses default TrackerCallback monitor value.
        mode (str): "auto", "min" or "max" to compare "monitor" values and define best model.
        input_type (str): "images" or None. Used to display sample predictions.
        validation_data (list): data used for sample predictions if input_type is set.
        predictions (int): number of predictions to make if input_type is set and validation_data is None.
        seed (int): initialize random generator for sample predictions if input_type is set and validation_data is None.
    """

    # Record if watch has been called previously (even in another instance)
    _watch_called = False

    def __init__(
        self,
        learn,
        log="gradients",
        save_model=True,
        monitor=None,
        mode="auto",
        input_type=None,
        validation_data=None,
        predictions=36,
        seed=12345,
    ):

        # Check if wandb.init has been called
        if wandb.run is None:
            raise ValueError("You must call wandb.init() before WandbCallback()")

        # Adapted from fast.ai "SaveModelCallback"
        if monitor is None:
            # use default TrackerCallback monitor value
            super().__init__(learn, mode=mode)
        else:
            super().__init__(learn, monitor=monitor, mode=mode)
        self.save_model = save_model
        self.model_path = Path(wandb.run.dir) / "bestmodel.pth"

        self.log = log
        self.input_type = input_type
        self.best = None

        # Select items for sample predictions to see evolution along training
        self.validation_data = validation_data
        if input_type and not self.validation_data:
            wandbRandom = random.Random(seed)  # For repeatability
            predictions = min(predictions, len(learn.data.valid_ds))
            indices = wandbRandom.sample(range(len(learn.data.valid_ds)), predictions)
            self.validation_data = [learn.data.valid_ds[i] for i in indices]

    def on_train_begin(self, **kwargs):
        "Call watch method to log model topology, gradients & weights"

        # Set self.best, method inherited from "TrackerCallback" by "SaveModelCallback"
        super().on_train_begin()

        # Ensure we don't call "watch" multiple times
        if not WandbCallback._watch_called:
            WandbCallback._watch_called = True

            # Logs model topology and optionally gradients and weights
            wandb.watch(self.learn.model, log=self.log)

    def on_epoch_end(self, epoch, smooth_loss, last_metrics, **kwargs):
        "Logs training loss, validation loss and custom metrics & log prediction samples & save model"

        if self.save_model:
            # Adapted from fast.ai "SaveModelCallback"
            current = self.get_monitor_value()
            if current is not None and self.operator(current, self.best):
                print(
                    "Better model found at epoch {} with {} value: {}.".format(
                        epoch, self.monitor, current
                    )
                )
                self.best = current

                # Save within wandb folder
                with self.model_path.open("wb") as model_file:
                    self.learn.save(model_file)

        # Log sample predictions if learn.predict is available
        if self.validation_data:
            try:
                self._wandb_log_predictions()
            except FastaiError as e:
                wandb.termwarn(e.message)
                self.validation_data = None  # prevent from trying again on next loop
            except Exception as e:
                wandb.termwarn("Unable to log prediction samples.\n{}".format(e))
                self.validation_data = None  # prevent from trying again on next loop

        # Log losses & metrics
        # Adapted from fast.ai "CSVLogger"
        logs = {
            name: stat
            for name, stat in list(
                zip(self.learn.recorder.names, [epoch, smooth_loss] + last_metrics)
            )
        }
        wandb.log(logs)

    def on_train_end(self, **kwargs):
        "Load the best model."

        if self.save_model:
            # Adapted from fast.ai "SaveModelCallback"
            if self.model_path.is_file():
                with self.model_path.open("rb") as model_file:
                    self.learn.load(model_file, purge=False)
                    print("Loaded best saved model from {}".format(self.model_path))

    def _wandb_log_predictions(self):
        "Log prediction samples"

        pred_log = []

        for x, y in self.validation_data:
            try:
                pred = self.learn.predict(x)
            except:
                raise FastaiError(
                    'Unable to run "predict" method from Learner to log prediction samples.'
                )

            # scalar -> likely to be a category
            # tensor of dim 1 -> likely to be multicategory
            if not pred[1].shape or pred[1].dim() == 1:
                pred_log.append(
                    wandb.Image(
                        x.data,
                        caption="Ground Truth: {}\nPrediction: {}".format(y, pred[0]),
                    )
                )

            # most vision datasets have a "show" function we can use
            elif hasattr(x, "show"):
                # log input data
                pred_log.append(wandb.Image(x.data, caption="Input data", grouping=3))

                # log label and prediction
                for im, capt in ((pred[0], "Prediction"), (y, "Ground Truth")):
                    # Resize plot to image resolution
                    # from https://stackoverflow.com/a/13714915
                    my_dpi = 100
                    fig = plt.figure(frameon=False, dpi=my_dpi)
                    h, w = x.size
                    fig.set_size_inches(w / my_dpi, h / my_dpi)
                    ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
                    ax.set_axis_off()
                    fig.add_axes(ax)

                    # Superpose label or prediction to input image
                    x.show(ax=ax, y=im)
                    pred_log.append(wandb.Image(fig, caption=capt))
                    plt.close(fig)

            # likely to be an image
            elif hasattr(y, "shape") and (
                (len(y.shape) == 2) or (len(y.shape) == 3 and y.shape[0] in [1, 3, 4])
            ):

                pred_log.extend(
                    [
                        wandb.Image(x.data, caption="Input data", grouping=3),
                        wandb.Image(pred[0].data, caption="Prediction"),
                        wandb.Image(y.data, caption="Ground Truth"),
                    ]
                )

            # we just log input data
            else:
                pred_log.append(wandb.Image(x.data, caption="Input data"))

            wandb.log({"Prediction Samples": pred_log}, commit=False)


class FastaiError(wandb.Error):
    pass



"""Batching file prepare requests to our API."""

import collections
import threading
import time
from six.moves import queue

# Request for a file to be prepared.
RequestPrepare = collections.namedtuple(
    "RequestPrepare", ("prepare_fn", "on_prepare", "response_queue")
)

RequestFinish = collections.namedtuple("RequestFinish", ())

ResponsePrepare = collections.namedtuple(
    "ResponsePrepare", ("upload_url", "upload_headers", "birth_artifact_id")
)


class StepPrepare(object):
    """A thread that batches requests to our file prepare API.

    Any number of threads may call prepare_async() in parallel. The PrepareBatcher thread
    will batch requests up and send them all to the backend at once.
    """

    def __init__(self, api, batch_time, inter_event_time, max_batch_size):
        self._api = api
        self._inter_event_time = inter_event_time
        self._batch_time = batch_time
        self._max_batch_size = max_batch_size
        self._request_queue = queue.Queue()
        self._thread = threading.Thread(target=self._thread_body)
        self._thread.daemon = True

    def _thread_body(self):
        while True:
            request = self._request_queue.get()
            if isinstance(request, RequestFinish):
                break
            finish, batch = self._gather_batch(request)
            prepare_response = self._prepare_batch(batch)
            # send responses
            for prepare_request in batch:
                name = prepare_request.prepare_fn()["name"]
                response_file = prepare_response[name]
                upload_url = response_file["uploadUrl"]
                upload_headers = response_file["uploadHeaders"]
                birth_artifact_id = response_file["artifact"]["id"]
                if prepare_request.on_prepare:
                    prepare_request.on_prepare(
                        upload_url, upload_headers, birth_artifact_id
                    )
                prepare_request.response_queue.put(
                    ResponsePrepare(upload_url, upload_headers, birth_artifact_id)
                )
            if finish:
                break

    def _gather_batch(self, first_request):
        batch_start_time = time.time()
        batch = [first_request]
        while True:
            try:
                request = self._request_queue.get(
                    block=True, timeout=self._inter_event_time
                )
                if isinstance(request, RequestFinish):
                    return True, batch
                batch.append(request)
                remaining_time = self._batch_time - (time.time() - batch_start_time)
                if remaining_time < 0 or len(batch) >= self._max_batch_size:
                    break
            except queue.Empty:
                break
        return False, batch

    def _prepare_batch(self, batch):
        """Execute the prepareFiles API call.

        Arguments:
            batch: List of RequestPrepare objects
        Returns:
            dict of (save_name: ResponseFile) pairs where ResponseFile is a dict with
                an uploadUrl key. The value of the uploadUrl key is None if the file
                already exists, or a url string if the file should be uploaded.
        """
        file_specs = []
        for prepare_request in batch:
            file_spec = prepare_request.prepare_fn()
            file_specs.append(file_spec)
        return self._api.create_artifact_files(file_specs)

    def prepare_async(self, prepare_fn, on_prepare=None):
        """Request the backend to prepare a file for upload.
        
        Returns:
            response_queue: a queue containing the prepare result. The prepare result is
                either a file upload url, or None if the file doesn't need to be uploaded.
        """
        response_queue = queue.Queue()
        self._request_queue.put(RequestPrepare(prepare_fn, on_prepare, response_queue))
        return response_queue

    def prepare(self, prepare_fn):
        return self.prepare_async(prepare_fn).get()

    def start(self):
        self._thread.start()

    def finish(self):
        self._request_queue.put(RequestFinish())

    def is_alive(self):
        return self._thread.is_alive()

    def shutdown(self):
        self.finish()
        self._thread.join()



#
# -*- coding: utf-8 -*-
"""
Log in to Weights & Biases, authenticating your machine to log data to your
account.
"""

from __future__ import print_function

import click
import wandb
from wandb.errors import UsageError

from .internal.internal_api import Api
from .lib import apikey
from .wandb_settings import Settings

if wandb.TYPE_CHECKING:  # type: ignore
    from typing import Dict, Optional  # noqa: F401 pylint: disable=unused-import


def login(anonymous=None, key=None, relogin=None, host=None, force=None):
    """
    Log in to W&B.

    Arguments:
        anonymous: (string, optional) Can be "must", "allow", or "never".
            If set to "must" we'll always login anonymously, if set to
            "allow" we'll only create an anonymous user if the user
            isn't already logged in.
        key: (string, optional) authentication key.
        relogin: (bool, optional) If true, will re-prompt for API key.
        host: (string, optional) The host to connect to.

    Returns:
        bool: if key is configured

    Raises:
        UsageError - if api_key can not configured and no tty
    """
    if wandb.setup()._settings._noop:
        return True
    kwargs = dict(locals())
    configured = _login(**kwargs)
    return True if configured else False


class _WandbLogin(object):
    def __init__(self):
        self.kwargs: Optional[Dict] = None
        self._settings: Optional[Settings] = None
        self._backend = None
        self._silent = None
        self._wl = None
        self._key = None

    def setup(self, kwargs):
        self.kwargs = kwargs

        # built up login settings
        login_settings: Settings = wandb.Settings()
        settings_param = kwargs.pop("_settings", None)
        if settings_param:
            login_settings._apply_settings(settings_param)
        _logger = wandb.setup()._get_logger()
        login_settings._apply_login(kwargs, _logger=_logger)

        # make sure they are applied globally
        self._wl = wandb.setup(settings=login_settings)
        self._settings = self._wl._settings

    def is_apikey_configured(self):
        return apikey.api_key(settings=self._settings) is not None

    def set_backend(self, backend):
        self._backend = backend

    def set_silent(self, silent):
        self._silent = silent

    def login(self):
        apikey_configured = self.is_apikey_configured()
        if self._settings.relogin:
            apikey_configured = False
        if not apikey_configured:
            return False

        if not self._silent:
            self.login_display()

        return apikey_configured

    def login_display(self):
        # check to see if we got an entity from the setup call
        active_entity = self._wl._get_entity()
        login_info_str = "(use `wandb login --relogin` to force relogin)"
        if active_entity:
            login_state_str = "Currently logged in as:"
            wandb.termlog(
                "{} {} {}".format(
                    login_state_str,
                    click.style(active_entity, fg="yellow"),
                    login_info_str,
                ),
                repeat=False,
            )
        else:
            login_state_str = "W&B API key is configured"
            wandb.termlog(
                "{} {}".format(login_state_str, login_info_str,), repeat=False,
            )

    def configure_api_key(self, key):
        if self._settings._jupyter and not self._settings._silent:
            wandb.termwarn(
                (
                    "If you're specifying your api key in code, ensure this "
                    "code is not shared publically.\nConsider setting the "
                    "WANDB_API_KEY environment variable, or running "
                    "`wandb login` from the command line."
                )
            )
        apikey.write_key(self._settings, key)
        self.update_session(key)
        self._key = key

    def update_session(self, key):
        _logger = wandb.setup()._get_logger()
        settings: Settings = wandb.Settings()
        login_settings = dict(api_key=key) if key else dict(mode="offline")
        settings._apply_source_login(login_settings, _logger=_logger)
        self._wl._update(settings=settings)
        # Whenever the key changes, make sure to pull in user settings
        # from server.
        if not self._wl.settings._offline:
            self._wl._update_user_settings()

    def prompt_api_key(self):
        api = Api(self._settings)
        key = apikey.prompt_api_key(
            self._settings,
            api=api,
            no_offline=self._settings.force,
            no_create=self._settings.force,
        )
        if key is False:
            directive = (
                "wandb login [your_api_key]"
                if self._settings._cli_only_mode
                else "wandb.login(key=[your_api_key])"
            )
            raise UsageError("api_key not configured (no-tty). call " + directive)

        self.update_session(key)
        self._key = key

    def propogate_login(self):
        # TODO(jhr): figure out if this is really necessary
        if self._backend:
            # TODO: calling this twice is gross, this deserves a refactor
            # Make sure our backend picks up the new creds
            # _ = self._backend.interface.communicate_login(key, anonymous)
            pass


def _login(
    anonymous=None,
    key=None,
    relogin=None,
    host=None,
    force=None,
    _backend=None,
    _silent=None,
    _disable_warning=None,
):
    kwargs = dict(locals())
    _disable_warning = kwargs.pop("_disable_warning", None)

    if wandb.run is not None:
        if not _disable_warning:
            wandb.termwarn("Calling wandb.login() after wandb.init() has no effect.")
        return True

    wlogin = _WandbLogin()

    _backend = kwargs.pop("_backend", None)
    if _backend:
        wlogin.set_backend(_backend)

    _silent = kwargs.pop("_silent", None)
    if _silent:
        wlogin.set_silent(_silent)

    # configure login object
    wlogin.setup(kwargs)

    if wlogin._settings._offline:
        return False
    elif wandb.util._is_kaggle() and not wandb.util._has_internet():
        wandb.termerror(
            "To use W&B in kaggle you must enable internet in the settings panel on the right."
        )
        return False

    # perform a login
    logged_in = wlogin.login()

    key = kwargs.get("key")
    if key:
        wlogin.configure_api_key(key)

    if logged_in:
        return logged_in

    if not key:
        wlogin.prompt_api_key()

    # make sure login credentials get to the backend
    wlogin.propogate_login()

    return wlogin._key or False



#
# -*- coding: utf-8 -*-
"""Writer thread."""

from __future__ import print_function

import logging

from . import datastore


logger = logging.getLogger(__name__)


class WriteManager(object):
    def __init__(
        self, settings, record_q, result_q,
    ):
        self._settings = settings
        self._record_q = record_q
        self._result_q = result_q
        self._ds = None

    def open(self):
        self._ds = datastore.DataStore()
        self._ds.open_for_write(self._settings.sync_file)

    def write(self, record):
        if not self._ds:
            self.open()

        record_type = record.WhichOneof("record_type")
        assert record_type

        self._ds.write(record)

    def finish(self):
        if self._ds:
            self._ds.close()



#
# -*- coding: utf-8 -*-
"""Backend - Send to internal process

Manage backend.

"""

import logging
import multiprocessing
import os
import sys
import threading

import wandb

from ..interface import interface
from ..internal.internal import wandb_internal

logger = logging.getLogger("wandb")


class BackendThread(threading.Thread):
    """Class to running internal process as a thread."""

    def __init__(self, target, kwargs) -> None:
        threading.Thread.__init__(self)
        self._target = target
        self._kwargs = kwargs
        self.daemon = True
        self.pid = 0

    def run(self):
        self._target(**self._kwargs)


class Backend(object):
    # multiprocessing context or module
    _multiprocessing: object

    def __init__(self, settings=None, log_level=None):
        self._done = False
        self.record_q = None
        self.result_q = None
        self.wandb_process = None
        self.interface = None
        self._internal_pid = None
        self._settings = settings
        self._log_level = log_level

        self._multiprocessing = multiprocessing
        self._multiprocessing_setup()

    def _hack_set_run(self, run):
        self.interface._hack_set_run(run)

    def _multiprocessing_setup(self):
        if self._settings.start_method == "thread":
            return

        # defaulting to spawn for now, fork needs more testing
        start_method = self._settings.start_method or "spawn"

        # TODO: use fork context if unix and frozen?
        # if py34+, else fall back
        if not hasattr(multiprocessing, "get_context"):
            return
        all_methods = multiprocessing.get_all_start_methods()
        logger.info(
            "multiprocessing start_methods={}, using: {}".format(
                ",".join(all_methods), start_method
            )
        )
        ctx = multiprocessing.get_context(start_method)
        self._multiprocessing = ctx

    def ensure_launched(self):
        """Launch backend worker if not running."""
        settings = dict(self._settings or ())
        settings["_log_level"] = self._log_level or logging.DEBUG

        # TODO: this is brittle and should likely be handled directly on the
        # settings object.  Multi-processing blows up when it can't pickle
        # objects.
        if "_early_logger" in settings:
            del settings["_early_logger"]

        self.record_q = self._multiprocessing.Queue()
        self.result_q = self._multiprocessing.Queue()
        if settings.get("start_method") != "thread":
            process_class = self._multiprocessing.Process
        else:
            process_class = BackendThread
            # disable interal process checks since we are one process
            wandb._set_internal_process(disable=True)
        self.wandb_process = process_class(
            target=wandb_internal,
            kwargs=dict(
                settings=settings, record_q=self.record_q, result_q=self.result_q,
            ),
        )
        self.wandb_process.name = "wandb_internal"

        # Support running code without a: __name__ == "__main__"
        save_mod_name = None
        save_mod_path = None
        main_module = sys.modules["__main__"]
        main_mod_spec = getattr(main_module, "__spec__", None)
        main_mod_path = getattr(main_module, "__file__", None)
        main_mod_name = None
        if main_mod_spec:
            main_mod_name = getattr(main_mod_spec, "name", None)
        if main_mod_name is not None:
            save_mod_name = main_mod_name
            main_module.__spec__.name = "wandb.mpmain"
        elif main_mod_path is not None:
            save_mod_path = main_module.__file__
            fname = os.path.join(
                os.path.dirname(wandb.__file__), "mpmain", "__main__.py"
            )
            main_module.__file__ = fname

        logger.info("starting backend process...")
        # Start the process with __name__ == "__main__" workarounds
        self.wandb_process.start()
        self._internal_pid = self.wandb_process.pid
        logger.info(
            "started backend process with pid: {}".format(self.wandb_process.pid)
        )

        # Undo temporary changes from: __name__ == "__main__"
        if save_mod_name:
            main_module.__spec__.name = save_mod_name
        elif save_mod_path:
            main_module.__file__ = save_mod_path

        self.interface = interface.BackendSender(
            process=self.wandb_process, record_q=self.record_q, result_q=self.result_q,
        )

    def server_connect(self):
        """Connect to server."""
        pass

    def server_status(self):
        """Report server status."""
        pass

    def cleanup(self):
        # TODO: make _done atomic
        if self._done:
            return
        self._done = True
        self.interface.join()
        self.wandb_process.join()
        self.record_q.close()
        self.result_q.close()
        # No printing allowed from here until redirect restore!!!



# File is generated by: tox -e codemod
import os
import logging
import tempfile as builtin_tempfile
import time
from six.moves import queue
import warnings

import wandb
import wandb.util
from wandb.compat import tempfile

from wandb.filesync import stats
from wandb.filesync import step_checksum
from wandb.filesync import step_upload


def resolve_path(path):
    try:
        from pathlib import Path

        return str(Path(path).resolve())
    except:
        # Pathlib isn't present for python versions earlier than 3.3
        return os.path.realpath(path)


# Get rid of cleanup warnings in Python 2.7.
warnings.filterwarnings(
    "ignore", "Implicitly cleaning up", RuntimeWarning, "wandb.compat.tempfile"
)


# Temporary directory for copies we make of some file types to
# reduce the probability that the file gets changed while we're
# uploading it.
TMP_DIR = tempfile.TemporaryDirectory("wandb")


logger = logging.getLogger(__file__)


class FilePusher(object):
    """Parallel file upload class.
    This manages uploading multiple files in parallel. It will restart a given file's
    upload job if it receives a notification that that file has been modified.
    The finish() method will block until all events have been processed and all
    uploads are complete.
    """

    MAX_UPLOAD_JOBS = 64

    def __init__(self, api, silent=False):
        self._api = api

        self._tempdir = tempfile.TemporaryDirectory("wandb")

        self._stats = stats.Stats()

        self._incoming_queue = queue.Queue()
        self._event_queue = queue.Queue()

        self._step_checksum = step_checksum.StepChecksum(
            self._api,
            self._tempdir,
            self._incoming_queue,
            self._event_queue,
            self._stats,
        )
        self._step_checksum.start()

        self._step_upload = step_upload.StepUpload(
            self._api,
            self._stats,
            self._event_queue,
            self.MAX_UPLOAD_JOBS,
            silent=silent,
        )
        self._step_upload.start()

        # Holds refs to tempfiles if users need to make a temporary file that
        # stays around long enough for file pusher to sync
        # TODO(artifacts): maybe don't do this
        self._temp_file_refs = []

    def get_status(self):
        running = self.is_alive()
        summary = self._stats.summary()
        return running, summary

    def print_status(self, prefix=True):
        step = 0
        spinner_states = ["-", "\\", "|", "/"]
        stop = False
        while True:
            if not self.is_alive():
                stop = True
            summary = self._stats.summary()
            line = " %.2fMB of %.2fMB uploaded (%.2fMB deduped)\r" % (
                summary["uploaded_bytes"] / 1048576.0,
                summary["total_bytes"] / 1048576.0,
                summary["deduped_bytes"] / 1048576.0,
            )
            line = spinner_states[step % 4] + line
            step += 1
            wandb.termlog(line, newline=False, prefix=prefix)
            if stop:
                break
            time.sleep(0.25)
        dedupe_fraction = (
            summary["deduped_bytes"] / float(summary["total_bytes"])
            if summary["total_bytes"] > 0
            else 0
        )
        if dedupe_fraction > 0.01:
            wandb.termlog(
                "W&B sync reduced upload amount by %.1f%%             "
                % (dedupe_fraction * 100),
                prefix=prefix,
            )
        # clear progress line.
        wandb.termlog(" " * 79, prefix=prefix)

    def file_counts_by_category(self):
        return self._stats.file_counts_by_category()

    def file_changed(
        self,
        save_name,
        path,
        artifact_id=None,
        copy=True,
        use_prepare_flow=False,
        save_fn=None,
        digest=None,
    ):
        """Tell the file pusher that a file's changed and should be uploaded.
        Arguments:
            save_name: string logical location of the file relative to the run
                directory.
            path: actual string path of the file to upload on the filesystem.
        """
        # Tests in linux were failing because wandb-events.jsonl didn't exist
        if not os.path.exists(path) or not os.path.isfile(path):
            return
        if os.path.getsize(path) == 0:
            return

        save_name = wandb.util.to_forward_slash_path(save_name)
        event = step_checksum.RequestUpload(
            path, save_name, artifact_id, copy, use_prepare_flow, save_fn, digest
        )
        self._incoming_queue.put(event)

    def store_manifest_files(self, manifest, artifact_id, save_fn):
        event = step_checksum.RequestStoreManifestFiles(manifest, artifact_id, save_fn)
        self._incoming_queue.put(event)

    def named_temp_file(self, mode="w+b"):
        # get a named temp file that the file pusher with hold a reference to so it
        # doesn't get gc'd. Obviously, we shouldn't do this very much :). It's currently
        # used for artifact metadata.
        f = builtin_tempfile.NamedTemporaryFile(mode=mode, delete=False)
        self._temp_file_refs.append(f)
        return f

    def commit_artifact(
        self, artifact_id, finalize=True, before_commit=None, on_commit=None
    ):
        event = step_checksum.RequestCommitArtifact(
            artifact_id, finalize, before_commit, on_commit
        )
        self._incoming_queue.put(event)

    def finish(self):
        logger.info("shutting down file pusher")
        self._incoming_queue.put(step_checksum.RequestFinish())

    def join(self):
        # NOTE: must have called finish before join
        logger.info("waiting for file pusher")
        while self.is_alive():
            time.sleep(0.5)

    def is_alive(self):
        return self._step_checksum.is_alive() or self._step_upload.is_alive()



from __future__ import absolute_import

import sys
import six
import argparse
import copy
import json
import os
import yaml
import importlib
import re

import wandb
from wandb import trigger
from wandb.util import add_import_hook


_import_hook = None
_run_once = False
_args_argparse = None
_args_system = None
_args_absl = None
_magic_init_seen = False
_magic_config = {}


class ArgumentException(Exception):
    pass


class SafeArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentException()


def _merge_dicts(source, destination):
    for key, value in source.items():
        if isinstance(value, dict):
            node = destination.setdefault(key, {})
            _merge_dicts(value, node)
        else:
            destination[key] = value
    return destination


def _dict_from_keyval(k, v, json_parse=True):
    d = ret = {}
    keys = k.split(".")
    for k in keys[:-1]:
        d = d.setdefault(k, {})
    if json_parse:
        try:
            v = json.loads(v.strip('"'))
        except ValueError:
            pass
    d[keys[-1]] = v
    return ret


def _magic_get_config(k, default):
    d = _magic_config
    keys = k.split(".")
    for k in keys[:-1]:
        d = d.get(k, {})
    return d.get(keys[-1], default)


_magic_defaults = {
    "enable": None,
    #'wandb': {
    #    'disable': None,
    # },
    "keras": {
        "fit": {
            "callbacks": {
                "tensorboard": {
                    "enable": True,
                    "duplicate": False,
                    "overwrite": False,
                    "write_graph": None,
                    "histogram_freq": None,
                    "update_freq": None,
                    "write_grads": None,
                    "write_images": None,
                    "batch_size": None,
                },
                "wandb": {
                    "enable": True,
                    "duplicate": False,
                    "overwrite": False,
                    "log_gradients": None,
                    "log_weights": None,
                    "data_type": "auto",
                    "input_type": None,
                    "output_type": None,
                    "log_evaluation": None,
                    "labels": None,
                    "predictions": None,
                    "save_model": None,
                    "save_weights_only": None,
                    "monitor": None,
                    "mode": None,
                    "verbose": None,
                },
                "epochs": None,
                "batch_size": None,
            }
        },
        #'compile': {
        #       'optimizer': {
        #           'name': False,
        #           },
        #       'loss': None,
        #    },
    },
    "args": {"absl": None, "argparse": None, "sys": None,},
}


def _parse_magic(val):
    # attempt to treat string as a json
    not_set = {}
    if val is None:
        return _magic_defaults, not_set
    if val.startswith("{"):
        try:
            val = json.loads(val)
        except ValueError:
            wandb.termwarn("Unable to parse magic json", repeat=False)
            return _magic_defaults, not_set
        conf = _merge_dicts(_magic_defaults, {})
        return _merge_dicts(val, conf), val
    if os.path.isfile(val):
        try:
            with open(val, "r") as stream:
                val = yaml.safe_load(stream)
        except IOError as e:
            wandb.termwarn("Unable to read magic config file", repeat=False)
            return _magic_defaults, not_set
        except yaml.YAMLError as e:
            wandb.termwarn("Unable to parse magic yaml file", repeat=False)
            return _magic_defaults, not_set
        conf = _merge_dicts(_magic_defaults, {})
        return _merge_dicts(val, conf), val
    # parse as a list of key value pairs
    if val.find("=") > 0:
        # split on commas but ignore commas inside quotes
        # Using this re allows env variable parsing like:
        # WANDB_MAGIC=key1='"["cat","dog","pizza"]"',key2=true
        items = re.findall(r'(?:[^\s,"]|"(?:\\.|[^"])*")+', val)
        conf_set = {}
        for kv in items:
            kv = kv.split("=")
            if len(kv) != 2:
                wandb.termwarn("Unable to parse magic key value pair", repeat=False)
                continue
            d = _dict_from_keyval(*kv)
            _merge_dicts(d, conf_set)
        conf = _merge_dicts(_magic_defaults, {})
        return _merge_dicts(conf_set, conf), conf_set
    wandb.termwarn("Unable to parse magic parameter", repeat=False)
    return _magic_defaults, not_set


def set_entity(value, env=None):
    if env is None:
        env = os.environ


def _fit_wrapper(self, fn, generator=None, *args, **kwargs):
    trigger.call("on_fit")
    keras = sys.modules.get("keras", None)
    tfkeras = sys.modules.get("tensorflow.python.keras", None)
    epochs = kwargs.pop("epochs", None)
    batch_size = kwargs.pop("batch_size", None)

    magic_epochs = _magic_get_config("keras.fit.epochs", None)
    if magic_epochs is not None:
        epochs = magic_epochs
    magic_batch_size = _magic_get_config("keras.fit.batch_size", None)
    if magic_batch_size is not None:
        batch_size = magic_batch_size
    callbacks = kwargs.pop("callbacks", [])

    tb_enabled = _magic_get_config("keras.fit.callbacks.tensorboard.enable", None)
    if tb_enabled:
        k = getattr(self, "_keras_or_tfkeras", None)
        if k:
            tb_duplicate = _magic_get_config(
                "keras.fit.callbacks.tensorboard.duplicate", None
            )
            tb_overwrite = _magic_get_config(
                "keras.fit.callbacks.tensorboard.overwrite", None
            )
            tb_present = any(
                [isinstance(cb, k.callbacks.TensorBoard) for cb in callbacks]
            )
            if tb_present and tb_overwrite:
                callbacks = [
                    cb
                    for cb in callbacks
                    if not isinstance(cb, k.callbacks.TensorBoard)
                ]
            if tb_overwrite or tb_duplicate or not tb_present:
                tb_callback_kwargs = {"log_dir": wandb.run.dir}
                cb_args = (
                    "write_graph",
                    "histogram_freq",
                    "update_freq",
                    "write_grads",
                    "write_images",
                    "batch_size",
                )
                for cb_arg in cb_args:
                    v = _magic_get_config(
                        "keras.fit.callbacks.tensorboard." + cb_arg, None
                    )
                    if v is not None:
                        tb_callback_kwargs[cb_arg] = v
                tb_callback = k.callbacks.TensorBoard(**tb_callback_kwargs)
                callbacks.append(tb_callback)

    wandb_enabled = _magic_get_config("keras.fit.callbacks.wandb.enable", None)
    if wandb_enabled:
        wandb_duplicate = _magic_get_config("keras.fit.callbacks.wandb.duplicate", None)
        wandb_overwrite = _magic_get_config("keras.fit.callbacks.wandb.overwrite", None)
        wandb_present = any(
            [isinstance(cb, wandb.keras.WandbCallback) for cb in callbacks]
        )
        if wandb_present and wandb_overwrite:
            callbacks = [
                cb for cb in callbacks if not isinstance(cb, wandb.keras.WandbCallback)
            ]
        if wandb_overwrite or wandb_duplicate or not wandb_present:
            wandb_callback_kwargs = {}
            log_gradients = _magic_get_config(
                "keras.fit.callbacks.wandb.log_gradients", None
            )
            if log_gradients and kwargs.get("x") and kwargs.get("y"):
                wandb_callback_kwargs["log_gradients"] = log_gradients
            cb_args = (
                "predictions",
                "log_weights",
                "data_type",
                "save_model",
                "save_weights_only",
                "monitor",
                "mode",
                "verbose",
                "input_type",
                "output_type",
                "log_evaluation",
                "labels",
            )
            for cb_arg in cb_args:
                v = _magic_get_config("keras.fit.callbacks.wandb." + cb_arg, None)
                if v is not None:
                    wandb_callback_kwargs[cb_arg] = v
            wandb_callback = wandb.keras.WandbCallback(**wandb_callback_kwargs)
            callbacks.append(wandb_callback)

    kwargs["callbacks"] = callbacks
    if epochs is not None:
        kwargs["epochs"] = epochs
    if batch_size is not None:
        kwargs["batch_size"] = batch_size
    if generator:
        return fn(generator, *args, **kwargs)
    return fn(*args, **kwargs)


# NOTE(jhr): need to spell out all useable args so that users who inspect can see args
def _magic_fit(
    self,
    x=None,
    y=None,
    batch_size=None,
    epochs=1,
    # FIXME: there is more
    # verbose=1,
    # callbacks=None,
    # validation_split=0.,
    # validation_data=None,
    # shuffle=True,
    # class_weight=None,
    # sample_weight=None,
    # initial_epoch=0,
    # steps_per_epoch=None,
    # validation_steps=None,
    # validation_freq=1,
    # max_queue_size=10,
    # workers=1,
    # use_multiprocessing=False,
    *args,
    **kwargs
):
    if hasattr(self, "_wandb_internal_model"):
        return self._fit(
            x=x, y=y, batch_size=batch_size, epochs=epochs, *args, **kwargs
        )
    return _fit_wrapper(
        self, self._fit, x=x, y=y, batch_size=batch_size, epochs=epochs, *args, **kwargs
    )


def _magic_fit_generator(
    self,
    generator,
    steps_per_epoch=None,
    epochs=1,
    # FIXME: there is more
    # verbose=1,
    # verbose=1,
    # callbacks=None,
    # validation_data=None,
    # validation_steps=None,
    # validation_freq=1,
    # class_weight=None,
    # max_queue_size=10,
    # workers=1,
    ##use_multiprocessing=False,
    # shuffle=True,
    # initial_epoch=0,
    *args,
    **kwargs
):
    return _fit_wrapper(
        self,
        self._fit_generator,
        generator=generator,
        steps_per_epoch=steps_per_epoch,
        epochs=epochs,
        *args,
        **kwargs
    )


def _monkey_tfkeras():
    from tensorflow import keras as tfkeras

    models = getattr(tfkeras, "models", None)
    if not models:
        return
    models.Model._keras_or_tfkeras = tfkeras
    if models.Model.fit == _magic_fit:
        return
    models.Model._fit = models.Model.fit
    models.Model.fit = _magic_fit
    models.Model._fit_generator = models.Model.fit_generator
    models.Model.fit_generator = _magic_fit_generator


def _monkey_absl():
    from absl import app as absl_app

    def _absl_callback():
        absl_flags = sys.modules.get("absl.flags")
        if not absl_flags:
            return
        _flags = getattr(absl_flags, "FLAGS", None)
        if not _flags:
            return
        _flags_as_dict = getattr(_flags, "flag_values_dict", None)
        if not _flags_as_dict:
            return
        _flags_module = getattr(_flags, "find_module_defining_flag", None)
        if not _flags_module:
            return
        flags_dict = {}
        for f, v in six.iteritems(_flags_as_dict()):
            m = _flags_module(f)
            if not m or m.startswith("absl."):
                continue
            flags_dict[f] = v
        global _args_absl
        _args_absl = flags_dict

    call_after_init = getattr(absl_app, "call_after_init", None)
    if not call_after_init:
        return
    call_after_init(_absl_callback)


def _process_system_args():
    global _args_system
    # try using argparse
    parser = SafeArgumentParser(add_help=False)
    for num, arg in enumerate(sys.argv):
        try:
            next_arg = sys.argv[num + 1]
        except IndexError:
            next_arg = ""
        if arg.startswith(("-", "--")) and not next_arg.startswith(("-", "--")):
            try:
                parser.add_argument(arg)
            except ValueError:
                pass
    try:
        parsed, unknown = parser.parse_known_args()
    except ArgumentException:
        pass
    else:
        _args_system = vars(parsed)


def _monkey_argparse():
    argparse._ArgumentParser = argparse.ArgumentParser

    def _install():
        argparse.ArgumentParser = MonitoredArgumentParser

    def _uninstall():
        argparse.ArgumentParser = argparse._ArgumentParser

    def monitored(self, args, unknown=None):
        global _args_argparse
        _args_argparse = copy.deepcopy(vars(args))

    class MonitoredArgumentParser(argparse._ArgumentParser):
        def __init__(self, *args, **kwargs):
            _uninstall()
            super(MonitoredArgumentParser, self).__init__(*args, **kwargs)
            _install()

        def parse_args(self, *args, **kwargs):
            args = super(MonitoredArgumentParser, self).parse_args(*args, **kwargs)
            return args

        def parse_known_args(self, *args, **kwargs):
            args, unknown = super(MonitoredArgumentParser, self).parse_known_args(
                *args, **kwargs
            )
            if self._callback:
                self._callback(args, unknown=unknown)
            return args, unknown

    _install()
    argparse.ArgumentParser._callback = monitored


def _magic_update_config():
    # if we already have config set, dont add anymore
    if wandb.run and wandb.run.config:
        c = wandb.run.config
        user_config = dict(c.items())
        if user_config:
            return
    if _magic_get_config("args.absl", None) is False:
        global _args_absl
        _args_absl = None
    if _magic_get_config("args.argparse", None) is False:
        global _args_argparse
        _args_argparse = None
    if _magic_get_config("args.sys", None) is False:
        global _args_system
        _args_system = None
    # prefer absl, then argparse values, fallback to parsed system args
    args = _args_absl or _args_argparse or _args_system
    if args and wandb.run and wandb.run.config:
        wandb.run.config.update(args)


def _magic_init(**kwargs):
    magic_arg = kwargs.get("magic", None)
    if magic_arg is not None and magic_arg is not False:
        global _magic_init_seen
        if _magic_init_seen and magic_arg is not True:
            wandb.termwarn(
                "wandb.init() magic argument ignored because wandb magic has already been initialized",
                repeat=False,
            )
        _magic_init_seen = True
    else:
        wandb.termwarn(
            "wandb.init() arguments ignored because wandb magic has already been initialized",
            repeat=False,
        )


def magic_install(init_args=None):
    if wandb.setup().settings._noop:
        return
    global _run_once
    if _run_once:
        return
    _run_once = True

    global _magic_config
    global _import_hook
    from wandb.integration.keras import WandbCallback  # add keras import hooks first

    # parse config early, before we have wandb.config overrides
    _magic_config, magic_set = _parse_magic(wandb.env.get_magic())

    # we are implicitly enabling magic
    if _magic_config.get("enable") is None:
        _magic_config["enable"] = True
        magic_set["enable"] = True

    # allow early config to disable magic
    if not _magic_config.get("enable"):
        return

    # process system args
    _process_system_args()
    # install argparse wrapper
    in_jupyter_or_ipython = wandb.wandb_sdk.lib.ipython._get_python_type != "python"
    if not in_jupyter_or_ipython:
        _monkey_argparse()

    # track init calls
    trigger.register("on_init", _magic_init)

    # if wandb.init has already been called, this call is ignored
    init_args = init_args or {}
    init_args["magic"] = True
    wandb.init(**init_args)

    # parse magic from wandb.config (from flattened to dict)
    magic_from_config = {}
    MAGIC_KEY = "wandb_magic"
    for k in wandb.config.keys():
        if not k.startswith(MAGIC_KEY + "."):
            continue
        d = _dict_from_keyval(k, wandb.config[k], json_parse=False)
        _merge_dicts(d, magic_from_config)
    magic_from_config = magic_from_config.get(MAGIC_KEY, {})
    _merge_dicts(magic_from_config, _magic_config)

    # allow late config to disable magic
    if not _magic_config.get("enable"):
        return

    # store magic_set into config
    if magic_set:
        wandb.config["magic"] = magic_set
        wandb.config.persist()

    # Monkey patch tf.keras
    if "tensorflow.python.keras" in sys.modules or "keras" in sys.modules:
        _monkey_tfkeras()

    # Always setup import hooks looking for keras or tf.keras
    add_import_hook(fullname="keras", on_import=_monkey_tfkeras)
    add_import_hook(fullname="tensorflow.python.keras", on_import=_monkey_tfkeras)

    if "absl.app" in sys.modules:
        _monkey_absl()
    else:
        add_import_hook(fullname="absl.app", on_import=_monkey_absl)

    # update wandb.config on fit or program finish
    trigger.register("on_fit", _magic_update_config)
    trigger.register("on_finished", _magic_update_config)



#!/usr/bin/env python

import wandb
import argparse
import sys
from wandb.sweeps.config import tune
from wandb.sweeps.config.hyperopt import hp
from wandb.sweeps.config.tune.schedulers import AsyncHyperBandScheduler
from wandb.sweeps.config.tune.suggest.hyperopt import HyperOptSearch
from datetime import datetime

# define command line options
parser = argparse.ArgumentParser()
parser.add_argument("--sweep", type=str, help="create sweep")
parser.add_argument("--file", type=str, help="save sweep config")
parser.add_argument("--controller", action="store_true", help="run local controller")
parser.add_argument("--create", action="store_true", default=True, help="create sweep")
parser.add_argument("--nocreate", action="store_false", dest="create", default=True, help="do not create sweep")
args, more_args = parser.parse_known_args()
if more_args:
    print("[WARNING] ignored extra arguments: {}".format(', '.join(more_args)))

# define param defaults
config_defaults = dict(
    epochs=12,
    learning_rate=12,
    width=1.2,
    height=3.4,
    activation="relu",
    )

# define sweeps
sweep_registry = dict(
    hyperopt=lambda:
        tune.run(
            parser.prog,
            search_alg=HyperOptSearch(
                dict(
                    width=hp.uniform("width", 0, 20),
                    height=hp.uniform("height", -100, 100),
                    activation=hp.choice("activation", ["relu", "tanh"])),
                metric="mean_loss",
                mode="min"),
            scheduler=AsyncHyperBandScheduler(
                metric="mean_loss",
                mode="min"),
            num_samples=10,
            ).set(
                name=datetime.now().strftime("hyperopt_%y%m%d%H%M"),
                settings=dict(
                    agent_command_config_args=False,
                    agent_report_interval=0,
                    ),
                )
    )


def train():
    run = wandb.init(config=config_defaults)
    shorten = dict(width="w", height="h", activation="a")
    clean = lambda x: '{:0.1f}'.format(x) if isinstance(x, float) else x
    run.name = "run:" + ','.join([
        '{}={}'.format(shorten.get(k), clean(v)) for k, v in dict(run.config).items() if k in shorten])
    run.save()
    conf = dict(wandb.config)
    value = conf.get("width") + conf.get("height")
    wandb.log(dict(mean_loss=value))


def main():
    if args.sweep:
        sweep_config = sweep_registry.get(args.sweep)
        if not sweep_config:
            print("[ERROR] can not find sweep: {}".format(args.sweep))
            sys.exit(1)
        if args.file:
            # construct config
            sweep_config = sweep_config()
            sweep_config.save(args.file)
        if args.create:
            sweep_id = wandb.sweep(sweep_config)
            if args.controller:
                sweep = wandb.controller(sweep_id)
                sweep.run()
        return
    train()


if __name__ == "__main__":
    main()



#
"""
runid util.
"""

import shortuuid  # type: ignore


def generate_id():
    # ~3t run ids (36**8)
    run_gen = shortuuid.ShortUUID(alphabet=list("0123456789abcdefghijklmnopqrstuvwxyz"))
    return run_gen.random(8)



from wandb.sweeps import hyperband_stopping as search
import numpy as np


def test_hyperband_min_iter_bands():
    hbet = search.HyperbandEarlyTerminate.init_from_min_iter(3, 3)
    assert (hbet.bands[:3] == [3, 9, 27])


def test_hyperband_max_iter_bands():
    hbet = search.HyperbandEarlyTerminate.init_from_max_iter(81, 3, 3)
    assert (hbet.bands[:3] == [3, 9, 27])


class Run(object):
    def __init__(self, name, state, history):
        self.name = name
        self.state = state
        self.history = history


def test_init_from_max_iter():
    et = search.HyperbandEarlyTerminate.init_from_max_iter(18, 3, 2)
    assert et.bands == [2, 6]


def test_single_run():
    et = search.HyperbandEarlyTerminate.init_from_max_iter(18, 3, 2)
    stopped, lines = et.stop_runs({'metric': {
        'name': 'loss',
        'goal': 'minimize',
    }},
        [Run('a', 'running',
             [{'loss': 10},
              {'loss': 9},
              {'loss': 8},
              {'loss': 7},
              {'loss': 6},
              {'loss': 5},
              {'loss': 4},
              {'loss': 3},
              {'loss': 2},
              {'loss': 1},
              ])]
    )
    assert stopped == []


def test_2runs_band1_stop():
    et = search.HyperbandEarlyTerminate.init_from_max_iter(18, 3, 2)
    stopped, lines = et.stop_runs({'metric': {
        'name': 'loss',
        'goal': 'minimize',
    }},
        [Run('a', 'running',
             [{'loss': 10},
              {'loss': 9},
              {'loss': 8},
              {'loss': 7},
              {'loss': 6},
              {'loss': 5},
              {'loss': 4},
              {'loss': 3},
              {'loss': 2},
              {'loss': 1},
              ]),
         Run('b', 'running',
             [{'loss': 10},
              {'loss': 10},
              {'loss': 10},
              ]),
         ]
    )
    assert stopped == ['b']


def test_2runs_band1_pass():
    et = search.HyperbandEarlyTerminate.init_from_max_iter(18, 3, 2)
    stopped, lines = et.stop_runs({'metric': {
        'name': 'loss',
        'goal': 'minimize',
    }},
        [Run('a', 'running',
             [{'loss': 10},
              {'loss': 9},
              {'loss': 8},
              {'loss': 7},
              {'loss': 6},
              {'loss': 5},
              {'loss': 4},
              {'loss': 3},
              {'loss': 2},
              {'loss': 1},
              ]),
         Run('b', 'running',
             [{'loss': 10},
              {'loss': 10},
              {'loss': 6},
              ]),
         ]
    )
    assert stopped == []


def test_skipped_steps():
    et = search.HyperbandEarlyTerminate.init_from_max_iter(18, 3, 2)
    et._load_metric_name_and_goal({'metric': {
        'name': 'loss',
        'goal': 'minimize'
    }})
    line = et._load_run_metric_history(
        Run('a', 'running',
            [{'loss': 10},
             {'a': 9},
             {'a': 8},
             {'a': 7},
             {'loss': 6},
             {'a': 5},
             {'a': 4},
             {'a': 3},
             {'a': 2},
             {'loss': 1},
             ]))
    assert line == [10, 6, 1]

def test_2runs_band1_stop_2():
    et = search.HyperbandEarlyTerminate.init_from_max_iter(5, 3, 2)
    stopped, lines = et.stop_runs({'metric': {
        'name': 'loss',
        'goal': 'minimize',
    }},
        [Run('a', 'stopped',
             [{'loss': 10},
              {'loss': 9},
              {'loss': 8},
              {'loss': 7},
              {'loss': 6},
              {'loss': 5},
              {'loss': 4},
              {'loss': 3},
              {'loss': 2},
              {'loss': 1},
              ]),
         Run('b', 'running',
             [{'loss': 10},
              {'loss': 10},
              {'loss': 10},
              ]),
         ]
    )
    assert stopped == ['b']

def test_5runs_band1_stop_2():
    et = search.HyperbandEarlyTerminate.init_from_max_iter(5, 2, 2)
    # bands are at 1 and 2
    stopped, lines = et.stop_runs({'metric': {
        'name': 'loss',
        'goal': 'minimize',
    }},
        [Run('a', 'stopped',  # This wont be stopped because already stopped
             [{'loss': 10},
              {'loss': 9},
              ]),
        Run('b', 'running',   # This should be stopped
             [{'loss': 10},
              {'loss': 10},
              ]),
        Run('c', 'running',   # This passes band 1 but not band 2
             [{'loss': 10},
              {'loss': 8},
              {'loss': 8},
              ]),
        Run('d', 'running',
             [{'loss': 10},
              {'loss': 7},
              {'loss': 7},
              ]),
        Run('e', 'finished',
             [{'loss': 10},
              {'loss': 6},
              {'loss': 6},
              ]),
         ]
    )
    assert stopped == ['b', 'c']

def test_5runs_band1_stop_2_1stnoband():
    et = search.HyperbandEarlyTerminate.init_from_max_iter(5, 2, 2)
    # bands are at 1 and 2
    stopped, lines = et.stop_runs({'metric': {
        'name': 'loss',
        'goal': 'minimize',
    }},
        [Run('a', 'running',  # This wont be stopped because not at band 1
             [{'loss': 10},
              ]),
        Run('b', 'running',   # This should be stopped
             [{'loss': 10},
              {'loss': 10},
              ]),
        Run('c', 'running',   # This passes band 1 but not band 2
             [{'loss': 10},
              {'loss': 8},
              {'loss': 8},
              ]),
        Run('d', 'running',
             [{'loss': 10},
              {'loss': 7},
              {'loss': 7},
              ]),
        Run('e', 'finished',
             [{'loss': 10},
              {'loss': 6},
              {'loss': 6},
              ]),
         ]
    )
    assert stopped == ['b', 'c']

def test_eta_3():
    et = search.HyperbandEarlyTerminate.init_from_max_iter(9, 3, 2)
    # bands are at 1 and 3, thresholds are 7 and 4
    stopped, lines = et.stop_runs({'metric': {
        'name': 'loss',
        'goal': 'minimize',
    }},
        [Run('a', 'stopped',  # This wont be stopped because already stopped
             [{'loss': 10},
              {'loss': 9},
              ]),
        Run('b', 'running',   # This should be stopped
             [{'loss': 10},
              {'loss': 10},
              ]),
        Run('c', 'running',   # This fails the first threeshold but snuck in so we wont kill
             [{'loss': 10},
              {'loss': 8},
              {'loss': 8},
              {'loss': 3},
              ]),
        Run('d', 'running',
             [{'loss': 10},
              {'loss': 7},
              {'loss': 7},
              {'loss': 4},
              ]),
        Run('e', 'running', # this passes band 1 but doesn't pass band 2
             [{'loss': 10},
              {'loss': 6},
              {'loss': 6},
              {'loss': 6},
              ]),
         ]
    )
    assert stopped == ['b', 'e']

def test_eta_3_max():
    et = search.HyperbandEarlyTerminate.init_from_max_iter(9, 3, 2)
    # bands are at 1 and 3, thresholds are 7 and 4
    stopped, lines = et.stop_runs({'metric': {
        'name': 'loss',
        'goal': 'maximize',
    }},
        [Run('a', 'stopped',  # This wont be stopped because already stopped
             [{'loss': -10},
              {'loss': -9},
              ]),
        Run('b', 'running',   # This should be stopped
             [{'loss': -10},
              {'loss': -10},
              ]),
        Run('c', 'running',   # This fails the first threeshold but snuck in so we wont kill
             [{'loss': -10},
              {'loss': -8},
              {'loss': -8},
              {'loss': -3},
              ]),
        Run('d', 'running',
             [{'loss': -10},
              {'loss': -7},
              {'loss': -7},
              {'loss': -4},
              ]),
        Run('e', 'running', # this passes band 1 but doesn't pass band 2
             [{'loss': -10},
              {'loss': -6},
              {'loss': -6},
              {'loss': -6},
              ]),
         ]
    )
    assert stopped == ['b', 'e']





#
"""
tensor b watcher.
"""

import logging
import os
import socket
import sys
import threading
import time

import six
from six.moves import queue
import wandb
from wandb import util
from wandb.viz import custom_chart_panel_config, CustomChart

from . import run as internal_run

if wandb.TYPE_CHECKING:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from ..interface.interface import BackendSender
        from .settings_static import SettingsStatic
        from typing import Dict, List, Optional
        from wandb.proto.wandb_internal_pb2 import RunRecord
        from six.moves.queue import PriorityQueue
        from tensorboard.compat.proto.event_pb2 import ProtoEvent
        from tensorboard.backend.event_processing.event_file_loader import (
            EventFileLoader,
        )

        HistoryDict = Dict[str, object]

# Give some time for tensorboard data to be flushed
SHUTDOWN_DELAY = 5
ERROR_DELAY = 5
REMOTE_FILE_TOKEN = "://"
logger = logging.getLogger(__name__)


def _link_and_save_file(
    path: str, base_path: str, interface: "BackendSender", settings: "SettingsStatic"
) -> None:
    # TODO(jhr): should this logic be merged with Run.save()
    files_dir = settings.files_dir
    file_name = os.path.relpath(path, base_path)
    abs_path = os.path.abspath(path)
    wandb_path = os.path.join(files_dir, file_name)
    util.mkdir_exists_ok(os.path.dirname(wandb_path))
    # We overwrite existing symlinks because namespaces can change in Tensorboard
    if os.path.islink(wandb_path) and abs_path != os.readlink(wandb_path):
        os.remove(wandb_path)
        os.symlink(abs_path, wandb_path)
    elif not os.path.exists(wandb_path):
        os.symlink(abs_path, wandb_path)
    # TODO(jhr): need to figure out policy, live/throttled?
    interface.publish_files(dict(files=[(file_name, "live")]))


def is_tfevents_file_created_by(path: str, hostname: str, start_time: float) -> bool:
    """Checks if a path is a tfevents file created by hostname.

    tensorboard tfevents filename format:
        https://github.com/tensorflow/tensorboard/blob/f3f26b46981da5bd46a5bb93fcf02d9eb7608bc1/tensorboard/summary/writer/event_file_writer.py#L81
    tensorflow tfevents fielname format:
        https://github.com/tensorflow/tensorflow/blob/8f597046dc30c14b5413813d02c0e0aed399c177/tensorflow/core/util/events_writer.cc#L68
    """
    if not path:
        raise ValueError("Path must be a nonempty string")
    basename = os.path.basename(path)
    if basename.endswith(".profile-empty"):
        return False
    fname_components = basename.split(".")
    try:
        tfevents_idx = fname_components.index("tfevents")
    except ValueError:
        return False
    # check the hostname, which may have dots
    for i, part in enumerate(hostname.split(".")):
        try:
            fname_component_part = fname_components[tfevents_idx + 2 + i]
        except IndexError:
            return False
        if part != fname_component_part:
            return False
    try:
        created_time = int(fname_components[tfevents_idx + 1])
    except (ValueError, IndexError):
        return False
    # Ensure that the file is newer then our start time, and that it was
    # created from the same hostname.
    # TODO: we should also check the PID (also contained in the tfevents
    #     filename). Can we assume that our parent pid is the user process
    #     that wrote these files?
    return created_time >= start_time  # noqa: W503


class TBWatcher(object):
    _logdirs: "Dict[str, TBDirWatcher]"
    _watcher_queue: "PriorityQueue"

    def __init__(
        self,
        settings: "SettingsStatic",
        run_proto: "RunRecord",
        interface: "BackendSender",
        force: bool = False,
    ) -> None:
        self._logdirs = {}
        self._consumer = None
        self._settings = settings
        self._interface = interface
        self._run_proto = run_proto
        self._force = force
        # TODO(jhr): do we need locking in this queue?
        self._watcher_queue = queue.PriorityQueue()
        wandb.tensorboard.reset_state()

    def _calculate_namespace(self, logdir: str, rootdir: str) -> "Optional[str]":
        namespace: "Optional[str]"
        dirs = list(self._logdirs) + [logdir]

        if os.path.isfile(logdir):
            filename = os.path.basename(logdir)
        else:
            filename = ""

        if rootdir == "":
            rootdir = util.to_forward_slash_path(
                os.path.dirname(os.path.commonprefix(dirs))
            )
            # Tensorboard loads all tfevents files in a directory and prepends
            # their values with the path. Passing namespace to log allows us
            # to nest the values in wandb
            # Note that we strip '/' instead of os.sep, because elsewhere we've
            # converted paths to forward slash.
            namespace = logdir.replace(filename, "").replace(rootdir, "").strip("/")
            # TODO: revisit this heuristic, it exists because we don't know the
            # root log directory until more than one tfevents file is written to
            if len(dirs) == 1 and namespace not in ["train", "validation"]:
                namespace = None
        else:
            namespace = logdir.replace(filename, "").replace(rootdir, "").strip("/")

        return namespace

    def add(self, logdir: str, save: bool, root_dir: str) -> None:
        logdir = util.to_forward_slash_path(logdir)
        root_dir = util.to_forward_slash_path(root_dir)
        if logdir in self._logdirs:
            return
        namespace = self._calculate_namespace(logdir, root_dir)
        # TODO(jhr): implement the deferred tbdirwatcher to find namespace

        if not self._consumer:
            self._consumer = TBEventConsumer(
                self, self._watcher_queue, self._run_proto, self._settings
            )
            self._consumer.start()

        tbdir_watcher = TBDirWatcher(
            self, logdir, save, namespace, self._watcher_queue, self._force
        )
        self._logdirs[logdir] = tbdir_watcher
        tbdir_watcher.start()

    def finish(self) -> None:
        for tbdirwatcher in six.itervalues(self._logdirs):
            tbdirwatcher.shutdown()
        for tbdirwatcher in six.itervalues(self._logdirs):
            tbdirwatcher.finish()
        if self._consumer:
            self._consumer.finish()


class TBDirWatcher(object):
    def __init__(
        self,
        tbwatcher: "TBWatcher",
        logdir: str,
        save: bool,
        namespace: "Optional[str]",
        queue: "PriorityQueue",
        force: bool = False,
    ) -> None:
        self.directory_watcher = util.get_module(
            "tensorboard.backend.event_processing.directory_watcher",
            required="Please install tensorboard package",
        )
        # self.event_file_loader = util.get_module(
        #     "tensorboard.backend.event_processing.event_file_loader",
        #     required="Please install tensorboard package",
        # )
        self.tf_compat = util.get_module(
            "tensorboard.compat", required="Please install tensorboard package"
        )
        self._tbwatcher = tbwatcher
        self._generator = self.directory_watcher.DirectoryWatcher(
            logdir, self._loader(save, namespace), self._is_our_tfevents_file
        )
        self._thread = threading.Thread(target=self._thread_body)
        self._first_event_timestamp = None
        self._shutdown = threading.Event()
        self._queue = queue
        self._file_version = None
        self._namespace = namespace
        self._logdir = logdir
        self._hostname = socket.gethostname()
        self._force = force

    def start(self) -> None:
        self._thread.start()

    def _is_our_tfevents_file(self, path: str) -> bool:
        """Checks if a path has been modified since launch and contains tfevents"""
        if not path:
            raise ValueError("Path must be a nonempty string")
        if self._force:
            return True
        path = self.tf_compat.tf.compat.as_str_any(path)
        return is_tfevents_file_created_by(
            path, self._hostname, self._tbwatcher._settings._start_time
        )

    def _loader(self, save: bool = True, namespace: str = None) -> "EventFileLoader":
        """Incredibly hacky class generator to optionally save / prefix tfevent files"""
        _loader_interface = self._tbwatcher._interface
        _loader_settings = self._tbwatcher._settings
        try:
            from tensorboard.backend.event_processing import event_file_loader
        except ImportError:
            raise Exception("Please install tensorboard package")

        class EventFileLoader(event_file_loader.EventFileLoader):
            def __init__(self, file_path: str) -> None:
                super(EventFileLoader, self).__init__(file_path)
                if save:
                    if REMOTE_FILE_TOKEN in file_path:
                        logger.warning(
                            "Not persisting remote tfevent file: %s", file_path
                        )
                    else:
                        # TODO: save plugins?
                        logdir = os.path.dirname(file_path)
                        parts = list(os.path.split(logdir))
                        if namespace and parts[-1] == namespace:
                            parts.pop()
                            logdir = os.path.join(*parts)
                        _link_and_save_file(
                            path=file_path,
                            base_path=logdir,
                            interface=_loader_interface,
                            settings=_loader_settings,
                        )

        return EventFileLoader

    def _process_events(self, shutdown_call: bool = False) -> None:
        try:
            for event in self._generator.Load():
                self.process_event(event)
        except (
            self.directory_watcher.DirectoryDeletedError,
            StopIteration,
            RuntimeError,
            OSError,
        ) as e:
            # When listing s3 the directory may not yet exist, or could be empty
            logger.debug("Encountered tensorboard directory watcher error: %s", e)
            if not self._shutdown.is_set() and not shutdown_call:
                time.sleep(ERROR_DELAY)

    def _thread_body(self) -> None:
        """Check for new events every second"""
        shutdown_time: "Optional[float]" = None
        while True:
            self._process_events()
            if self._shutdown.is_set():
                now = time.time()
                if not shutdown_time:
                    shutdown_time = now + SHUTDOWN_DELAY
                elif now > shutdown_time:
                    break
            time.sleep(1)

    def process_event(self, event: "ProtoEvent") -> None:
        # print("\nEVENT:::", self._logdir, self._namespace, event, "\n")
        if self._first_event_timestamp is None:
            self._first_event_timestamp = event.wall_time

        if event.HasField("file_version"):
            self._file_version = event.file_version

        if event.HasField("summary"):
            self._queue.put(Event(event, self._namespace))

    def shutdown(self) -> None:
        self._process_events(shutdown_call=True)
        self._shutdown.set()

    def finish(self) -> None:
        self.shutdown()
        self._thread.join()


class Event(object):
    """An event wrapper to enable priority queueing"""

    def __init__(self, event: "ProtoEvent", namespace: "Optional[str]"):
        self.event = event
        self.namespace = namespace
        self.created_at = time.time()

    def __lt__(self, other: "Event") -> bool:
        if self.event.wall_time < other.event.wall_time:
            return True
        return False


class TBEventConsumer(object):
    """Consumes tfevents from a priority queue.  There should always
    only be one of these per run_manager.  We wait for 10 seconds of queued
    events to reduce the chance of multiple tfevent files triggering
    out of order steps.
    """

    def __init__(
        self,
        tbwatcher: TBWatcher,
        queue: "PriorityQueue",
        run_proto: "RunRecord",
        settings: "SettingsStatic",
        delay: int = 10,
    ) -> None:
        self._tbwatcher = tbwatcher
        self._queue = queue
        self._thread = threading.Thread(target=self._thread_body)
        self._shutdown = threading.Event()
        self.tb_history = TBHistory()
        self._delay = delay

        # This is a bit of a hack to get file saving to work as it does in the user
        # process. Since we don't have a real run object, we have to define the
        # datatypes callback ourselves.
        def datatypes_cb(fname: str) -> None:
            files = dict(files=[(fname, "now")])
            self._tbwatcher._interface.publish_files(files)

        self._internal_run = internal_run.InternalRun(run_proto, settings, datatypes_cb)

    def start(self) -> None:
        self._start_time = time.time()
        self._thread.start()

    def finish(self) -> None:
        self._delay = 0
        self._shutdown.set()
        try:
            event = self._queue.get(True, 1)
        except queue.Empty:
            event = None
        if event:
            self._handle_event(event, history=self.tb_history)
            items = self.tb_history._get_and_reset()
            for item in items:
                self._save_row(item,)
        self._thread.join()

    def _thread_body(self) -> None:
        while True:
            try:
                event = self._queue.get(True, 1)
                # Wait self._delay seconds from consumer start before logging events
                if (
                    time.time() < self._start_time + self._delay
                    and not self._shutdown.is_set()
                ):
                    self._queue.put(event)
                    time.sleep(0.1)
                    continue
            except queue.Empty:
                event = None
                if self._shutdown.is_set():
                    break
            if event:
                self._handle_event(event, history=self.tb_history)
                items = self.tb_history._get_and_reset()
                for item in items:
                    self._save_row(item,)
        # flush uncommitted data
        self.tb_history._flush()
        items = self.tb_history._get_and_reset()
        for item in items:
            self._save_row(item)

    def _handle_event(self, event: "ProtoEvent", history: "TBHistory" = None) -> None:
        wandb.tensorboard.log(
            event.event,
            step=event.event.step,
            namespace=event.namespace,
            history=history,
        )

    def _save_row(self, row: "HistoryDict") -> None:
        chart_keys = []
        for key, item in row.items():
            if isinstance(item, CustomChart):
                panel_config = custom_chart_panel_config(item, key, key + "_table")
                config = {"panel_type": "Vega2", "panel_config": panel_config}
                chart_keys.append(key)
                self._tbwatcher._interface.publish_config(
                    val=config, key=("_wandb", "visualize", key)
                )
                row[key] = item.table

        for chart_key in chart_keys:
            table = row[chart_key]
            row.pop(chart_key)
            row[chart_key + "_table"] = table

        self._tbwatcher._interface.publish_history(
            row, run=self._internal_run, publish_step=False
        )


class TBHistory(object):
    _data: "HistoryDict"
    _added: "List[HistoryDict]"

    def __init__(self) -> None:
        self._step = 0
        self._step_size = 0
        self._data = dict()
        self._added = []

    def _flush(self) -> None:
        if not self._data:
            return
        # A single tensorboard step may have too much data
        # we just drop the largest keys in the step if it does.
        # TODO: we could flush the data across multiple steps
        if self._step_size > util.MAX_LINE_SIZE:
            metrics = [(k, sys.getsizeof(v)) for k, v in self._data.items()]
            metrics.sort(key=lambda t: t[1], reverse=True)
            bad = 0
            dropped_keys = []
            for k, v in metrics:
                # TODO: (cvp) Added a buffer of 100KiB, this feels rather brittle.
                if self._step_size - bad < util.MAX_LINE_SIZE - 100000:
                    break
                else:
                    bad += v
                    dropped_keys.append(k)
                    del self._data[k]
            wandb.termwarn(
                "Step {} exceeds max data limit, dropping {} of the largest keys:".format(
                    self._step, len(dropped_keys)
                )
            )
            print("\t" + ("\n\t".join(dropped_keys)))
        self._data["_step"] = self._step
        self._added.append(self._data)
        self._step += 1
        self._step_size = 0

    def add(self, d: "HistoryDict") -> None:
        self._flush()
        self._data = dict()
        self._data.update(self._track_history_dict(d))

    def _track_history_dict(self, d: "HistoryDict") -> "HistoryDict":
        e = {}
        for k in d.keys():
            e[k] = d[k]
            self._step_size += sys.getsizeof(e[k])
        return e

    def _row_update(self, d: "HistoryDict") -> None:
        self._data.update(self._track_history_dict(d))

    def _get_and_reset(self) -> "List[HistoryDict]":
        added = self._added[:]
        self._added = []
        return added



"""
Tool for creating styles from a dictionary.

This is very similar to the Pygments style dictionary, with some additions:
- Support for reverse and blink.
- Support for ANSI color names. (These will map directly to the 16 terminal
  colors.)
"""
from six.moves.collections_abc import Mapping

from .base import Style, DEFAULT_ATTRS, ANSI_COLOR_NAMES
from .defaults import DEFAULT_STYLE_EXTENSIONS
from .utils import merge_attrs, split_token_in_parts
from six.moves import range

__all__ = (
    'style_from_dict',
)


def _colorformat(text):
    """
    Parse/validate color format.

    Like in Pygments, but also support the ANSI color names.
    (These will map to the colors of the 16 color palette.)
    """
    if text[0:1] == '#':
        col = text[1:]
        if col in ANSI_COLOR_NAMES:
            return col
        elif len(col) == 6:
            return col
        elif len(col) == 3:
            return col[0]*2 + col[1]*2 + col[2]*2
    elif text == '':
        return text

    raise ValueError('Wrong color format %r' % text)


def style_from_dict(style_dict, include_defaults=True):
    """
    Create a ``Style`` instance from a dictionary or other mapping.

    The dictionary is equivalent to the ``Style.styles`` dictionary from
    pygments, with a few additions: it supports 'reverse' and 'blink'.

    Usage::

        style_from_dict({
            Token: '#ff0000 bold underline',
            Token.Title: 'blink',
            Token.SomethingElse: 'reverse',
        })

    :param include_defaults: Include the defaults (built-in) styling for
        selected text, etc...)
    """
    assert isinstance(style_dict, Mapping)

    if include_defaults:
        s2 = {}
        s2.update(DEFAULT_STYLE_EXTENSIONS)
        s2.update(style_dict)
        style_dict = s2

    # Expand token inheritance and turn style description into Attrs.
    token_to_attrs = {}

    # (Loop through the tokens in order. Sorting makes sure that
    # we process the parent first.)
    for ttype, styledef in sorted(style_dict.items()):
        # Start from parent Attrs or default Attrs.
        attrs = DEFAULT_ATTRS

        if 'noinherit' not in styledef:
            for i in range(1, len(ttype) + 1):
                try:
                    attrs = token_to_attrs[ttype[:-i]]
                except KeyError:
                    pass
                else:
                    break

        # Now update with the given attributes.
        for part in styledef.split():
            if part == 'noinherit':
                pass
            elif part == 'bold':
                attrs = attrs._replace(bold=True)
            elif part == 'nobold':
                attrs = attrs._replace(bold=False)
            elif part == 'italic':
                attrs = attrs._replace(italic=True)
            elif part == 'noitalic':
                attrs = attrs._replace(italic=False)
            elif part == 'underline':
                attrs = attrs._replace(underline=True)
            elif part == 'nounderline':
                attrs = attrs._replace(underline=False)

            # prompt_toolkit extensions. Not in Pygments.
            elif part == 'blink':
                attrs = attrs._replace(blink=True)
            elif part == 'noblink':
                attrs = attrs._replace(blink=False)
            elif part == 'reverse':
                attrs = attrs._replace(reverse=True)
            elif part == 'noreverse':
                attrs = attrs._replace(reverse=False)

            # Pygments properties that we ignore.
            elif part in ('roman', 'sans', 'mono'):
                pass
            elif part.startswith('border:'):
                pass

            # Colors.

            elif part.startswith('bg:'):
                attrs = attrs._replace(bgcolor=_colorformat(part[3:]))
            else:
                attrs = attrs._replace(color=_colorformat(part))

        token_to_attrs[ttype] = attrs

    return _StyleFromDict(token_to_attrs)


class _StyleFromDict(Style):
    """
    Turn a dictionary that maps `Token` to `Attrs` into a style class.

    :param token_to_attrs: Dictionary that maps `Token` to `Attrs`.
    """
    def __init__(self, token_to_attrs):
        self.token_to_attrs = token_to_attrs

    def get_attrs_for_token(self, token):
        # Split Token.
        list_of_attrs = []
        for token in split_token_in_parts(token):
            list_of_attrs.append(self.token_to_attrs.get(token, DEFAULT_ATTRS))
        return merge_attrs(list_of_attrs)

    def invalidation_hash(self):
        return id(self.token_to_attrs)



# -*- coding: utf-8 -*-
"""Agent - Agent object.

Manage wandb agent.

"""

from __future__ import print_function

import sys
import click

import time

import os
from wandb import Settings
from wandb.apis import internal_runqueue
import subprocess
import json


class Agent(object):
    def __init__(self, spec):
        self._spec = spec
        # glob_config = os.path.expanduser('~/.config/wandb/settings')
        # loc_config = 'wandb/settings'
        # files = (glob_config, loc_config)
        settings = Settings()
        self._api = internal_runqueue.Api()
        self._settings = settings

    def check_queue(self):
        entity, project = self._spec.split("/")
        # ups = self._api.pop_from_run_queue(entity="jeff", project="super-agent")
        # print("ent", entity, project)
        try:
            ups = self._api.pop_from_run_queue(entity=entity, project=project)
        except Exception as e:
            print("nothing to run.")
            return None
        return ups

    def run_job(self, job):
        print("agent: got job", job)
        # j = json.loads(job)
        j = job
        cl = j.get("runSpec", {}).get("input", {}).get("config", {})
        c = cl[0]
        # print("got config", c)
        outjson = json.dumps(c)

        with open("config.json", "w") as f:
            print(outjson, file=f)

        command = ["python", "train.py"]
        env = os.environ

        kwargs = dict()
        popen = subprocess.Popen(command, env=env, **kwargs)
        popen.wait()

    def loop(self):
        while True:
            job = self.check_queue()
            if not job:
                time.sleep(20)
                continue
            self.run_job(job)
            time.sleep(5)


def run_agent(spec):
    if not spec or len(spec) != 1 or len(spec[0].split("/")) != 2:
        click.echo("ERROR: Specify agent spec in the form: 'entity/project'")
        sys.exit(1)
    spec = spec[0]
    print("Super agent spec: %s" % spec)

    agent = Agent(spec)

    agent.loop()



import wandb
from wandb import util

chart_limit = wandb.Table.MAX_ROWS


def confusion_matrix(probs=None, y_true=None, preds=None, class_names=None, title=None):
    """
    Computes a multi-run confusion matrix.

    Arguments:
        probs (2-d arr): Shape [n_examples, n_classes]
        y_true (arr): Array of label indices.
        preds (arr): Array of predicted label indices.
        class_names (arr): Array of class names.

    Returns:
        Nothing. To see plots, go to your W&B run page then expand the 'media' tab
        under 'auto visualizations'.

    Example:
        ```
        vals = np.random.uniform(size=(10, 5))
        probs = np.exp(vals)/np.sum(np.exp(vals), keepdims=True, axis=1)
        y_true = np.random.randint(0, 5, size=(10))
        labels = ["Cat", "Dog", "Bird", "Fish", "Horse"]
        wandb.log({'confusion_matrix': wandb.plot.confusion_matrix(probs, y_true=y_true, class_names=labels)})
        ```
    """

    np = util.get_module(
        "numpy",
        required="confusion matrix requires the numpy library, install with `pip install numpy`",
    )
    # change warning
    assert probs is None or len(probs.shape) == 2, (
        "confusion_matrix has been updated to accept"
        " probabilities as the default first argument. Use preds=..."
    )

    assert (probs is None or preds is None) and not (
        probs is None and preds is None
    ), "Must provide probabilties or predictions but not both to confusion matrix"

    if probs is not None:
        preds = np.argmax(probs, axis=1).tolist()

    assert len(preds) == len(
        y_true
    ), "Number of predictions and label indices must match"

    if class_names is not None:
        n_classes = len(class_names)
        class_inds = [i for i in range(n_classes)]
        assert max(preds) <= len(
            class_names
        ), "Higher predicted index than number of classes"
        assert max(y_true) <= len(
            class_names
        ), "Higher label class index than number of classes"
    else:
        class_inds = set(preds).union(set(y_true))
        n_classes = len(class_inds)
        class_names = ["Class_{}".format(i) for i in range(1, n_classes + 1)]

    # get mapping of inds to class index in case user has weird prediction indices
    class_mapping = {}
    for i, val in enumerate(sorted(list(class_inds))):
        class_mapping[val] = i
    counts = np.zeros((n_classes, n_classes))
    for i in range(len(preds)):
        counts[class_mapping[y_true[i]], class_mapping[preds[i]]] += 1

    data = []
    for i in range(n_classes):
        for j in range(n_classes):
            data.append([class_names[i], class_names[j], counts[i, j]])

    fields = {
        "Actual": "Actual",
        "Predicted": "Predicted",
        "nPredictions": "nPredictions",
    }
    title = title or ""
    return wandb.plot_table(
        "wandb/confusion_matrix/v1",
        wandb.Table(columns=["Actual", "Predicted", "nPredictions"], data=data),
        fields,
        {"title": title},
    )
